
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
*/

/*!
 * \file
 *                Multiscale KSVD Denoising
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File algo.h
 * \brief Contains the abstract class Algo */

#ifndef ALGO_H
#define ALGO_H

/// Default shift overlap 
#define OVERLAP_FACTOR 1
#define OVERLAP_FACTOR_VIDEO 1
#define DEFAULT_LOW_SPARSITY 3
#define DEFAULT_MAX_SPARSITY_INPAINTING 20
#define DEFAULT_MAX_SPARSITY_INPAINTING_VIDEO 50

#include "dictionary.h"
#include "video.h"
#include "processor.h"
#include <qdir.h>
#ifdef _OPENMP
#include <omp.h>
#endif

/// Standard Template Library
using namespace std;

namespace KSVD {

   /*!
    * \brief Abstract class Algo. 
    *
    * This class is abstract. Every 
    * K-SVD algorithm should derive from this one and implement
    * the function exec(). The template parameter T is the type 
    * of the data, the parameter V is the number of channels,
    * 1 for GRAY, 3 for RGB and any other value for a future
    * implementation. */
   template <typename T = double, int V = 1> class Algo {
      public:

         /// Constructor, original is optional
         /// input is a noisy image, if the original image 
         /// is provided, it will provides the PSNR values
         /// during the process
         Algo(const Image<T,V>* const input, 
               const QString& fileDictionary, 
               const Image<T,V>* const original = NULL,
               const int overlapFactor = OVERLAP_FACTOR);
         Algo(const Image<T,V>* const input, 
               Dictionary<T,V>* dictionary, 
               const Image<T,V>* const original = NULL,
               const int overlapFactor = OVERLAP_FACTOR);

         /// Virtual Destructor
         virtual ~Algo();

         /// execute the algorithm virtual method.
         /// This should be implemented in the derived classes
         virtual void exec(const T gam, const bool weightConstant) = 0; 
         /// print some information about the algorithm
         virtual void printInformation() const;
         /// print the PSNR between the original and the output
         void printPSNR() const;
         /// return the PSNR
         T inline PSNR() const;
         /// write the dictionary in an image
         void inline writeDictionaryToImage(const QString& file) const;
         /// Accessor for the filename of the output.
         bool saveFile(const QString& fileName) const;

         /// Accessor for the output
         inline const Image<T,V>& output() const {
            return *_output;
         };
         /// Accessor to the dictionary
         inline const Dictionary<T,V>& dictionary() const {
            return *_dictionary;
         };
         /// Modifier
         void inline setStats(bool st) { _stats = st;};
         inline void setDictionaryNull() { _dictionary = NULL; };
         void inline setN(int N) { _N = N;};

      protected:
         /// generate the output image from the data
         /// in the dictionary, which contains the sparse
         /// decomposition of each patch. \f$ \lambda \f$ is the 
         /// lagrange multiplier. beta is eventually a weight matrix
         void generateOutput(list<int>& listPatches,
               double lambda = 0.0, const Image<T,V>* const beta = NULL);
         /// generate a list of patches, each number in this list
         /// corresponds to one coordinate in the image
         int generateListPatches(list<int>& listPatches) const;
         /// handle the scheduling of the approximation of the patches
         /// If a compiler which supports OpenMP, the implementation
         /// is parallel.
         virtual void approximation(Processor<T,V>** processors,
               list<int>& listPatches, const int num_threads = 1,
               const bool weightConstant = true,
               const int offsetCount = 0);

         /// result of the algorithm
         Image<T,V>* _output;   
         /// input of the algorithm
         const Image<T,V>* const _input;    
         /// optional : original data.
         const Image<T,V>* const _original; 
         /// number of scales
         int _N;      
         /// dictionary
         Dictionary<T,V>* _dictionary;
         /// number of pixels between patches
         int _overlapFactor;
         /// size of the patches which define the overlapping level
         int _basePatch;
         /// print some statistics or not
         bool _stats;

      private:
         /// Copy constructor is forbidden
         explicit Algo(const Algo<T,V>& alg);
         /// operator= is forbidden
         Algo<T,V>& operator=(const Algo<T,V>& alg);
   };


   /// Constructor, original is optional
   /// input is a noisy image, if the original image 
   /// is provided, it will provides the PSNR values
   /// during the process
   /// TODO : factorize the code
   template <typename T, int V>
      Algo<T,V>::Algo(const Image<T,V>* const input, 
            const QString& fileDictionary, 
            const Image<T,V>* const original,
            const int overlapFactor) :
         _output(NULL), _input(input), _original(original), 
         _stats(true),
         _overlapFactor(overlapFactor) {
            _dictionary = new Dictionary<T,V>(fileDictionary);
            _N = _dictionary->N();
            if (_N > MAX_N) {
               cerr << "Error : N too big" << endl;
               exit(1);
            }
            _basePatch = static_cast<int>(static_cast<double>(
                     _dictionary->sizeEdgeBigPatches()));
         };

   template <typename T, int V>
      Algo<T,V>::Algo(const Image<T,V>* const input, 
            Dictionary<T,V>* dictionary, 
            const Image<T,V>* const original,
            const int overlapFactor) :
         _output(NULL), _input(input), _original(original), 
         _stats(true),
         _dictionary(dictionary),
         _overlapFactor(overlapFactor) {
            _N = _dictionary->N();
            if (_N > MAX_N) {
               cerr << "Error : N too big" << endl;
               exit(1);
            }
            _basePatch = static_cast<int>(static_cast<double>(
                     _dictionary->sizeEdgeBigPatches()));
         };

   /// Destructor
   template <typename T, int V>
      Algo<T,V>::~Algo() {
         delete(_dictionary);
         _dictionary = NULL;
         delete(_output);
         _output = NULL;
      };

   /// print some information about the algorithm
   template <typename T, int V>
      void Algo<T,V>::printInformation() const {
         cout << "The number of scales is " << _N << endl;
         cout << "The input image is : ";
         _input->printInformation();
         _dictionary->printInformation();
      };

   /// print the PSNR between the input and the output
   template<typename T, int V>
      void Algo<T,V>::printPSNR() const {
         if (!_output || !_original) return;
         cout << "PSNR : " << PSNR() << endl; 
         if (_original) {
            cout << "PSNR initial : " << _input->PSNR(*_original) << endl; 
         }
      };

   /// compute the PSNR
   template<typename T, int V>
      T inline Algo<T,V>::PSNR() const {
         assert(_original != NULL && _output != NULL);
         return _output->PSNR(*_original);
      };

   /// write the dictionary in an image
   template<typename T, int V>
      void inline Algo<T,V>::writeDictionaryToImage(
            const QString& file) const {
         _dictionary->writeDictionaryToImage(file);
         _dictionary->writeDictionary(file+".dict");
      };

   /// Accessor for output
   template<typename T, int V>
      bool Algo<T,V>::saveFile(const QString& fileName) const {
         bool res = _output && _output->saveFile(fileName);
         QString endfix("_input");
         QString file(fileName);
         if (_original)
            _original->saveFile(fileName+"_orig.png");
         return res && _input && _input->saveFile(
               file.insert(fileName.length()-4,endfix));
      };

   /// generate the output image from the data
   /// in the dictionary, which contains the sparse
   /// decomposition of each patch. \f$ \lambda \f$ is the 
   /// lagrange multiplier
   template<typename T, int V>
      void Algo<T,V>::generateOutput(
            list<int>& listPatches, double lambda,
            const Image<T,V>* const beta) {
         if (_output == NULL) {
            _output = new Image<T,V>(*_input);
         } else {
            *_output = *_input;
         }
         Image<T,V>* weights = 
            new Image<T,V>(_input->width(),_input->height());
         weights->clear();
         Image<T,V>* weightsAux = 
            new Image<T,V>(_input->width(),_input->height());
         *weightsAux = 1.0;
         int sizeEdgePatches = twoPow[_N-1]*
            _dictionary->sizeEdgeSmallPatches();
         for (list<int>::iterator it = listPatches.begin();
               it != listPatches.end();
               ++it) {
            weights->incrPatch(*it,sizeEdgePatches);
         }
         int numberPatchesPerPixel = 
            _dictionary->sizeEdgeBigPatches()*
            _dictionary->sizeEdgeBigPatches()/
            (_overlapFactor*_overlapFactor);
         (*weightsAux) *= lambda/(numberPatchesPerPixel);
         (*weightsAux) *= (*weights);
         if (beta != NULL) (*weightsAux) *= (*beta);
         (*_output) *= (*weightsAux);
         (*weights) += (*weightsAux);
         int count = 0;
         for (list<int>::iterator it = listPatches.begin();
               it != listPatches.end();
               ++it) {
            _dictionary->addPatch(*_output, *it, count,
                  sizeEdgePatches,_N);
            ++count;
         }
         (*_output) /= (*weights);
         _output->clipData();
         delete(weights);
         delete(weightsAux);
      };

   /// generate a list of patches, each number in this list
   /// corresponds to one coordinate in the image
   template<typename T, int V>
      int Algo<T,V>::generateListPatches(
            list<int>& listPatches) const {
         int count = 0;
         listPatches.clear();
         int sizeEdgePatches = twoPow[_N-1]*
            _dictionary->sizeEdgeSmallPatches();
         for (int i = 0; i<= _input->height()-sizeEdgePatches;
               i += _overlapFactor) {
            int offset = i*_input->width();
            for (int j = 0; j<= _input->width()-sizeEdgePatches;
                  j += _overlapFactor) {
               listPatches.push_back(offset+j);
               ++count;
               if ((j % _basePatch) == _basePatch-1 &&
                     j != _input->width()-sizeEdgePatches-1) {
                  j = MAX(j,MIN(j+sizeEdgePatches-_basePatch,
                           _input->width()-sizeEdgePatches-1-_basePatch));
               }
            }
            if ((i % _basePatch) == _basePatch-1 &&
                  i != _input->height()-sizeEdgePatches-1) {
               i = MAX(i,MIN(i+sizeEdgePatches-_basePatch,
                        _input->height()-sizeEdgePatches-1-_basePatch));
            }
         }
         cout << "Number of patches processed : " << count << endl;
         return count;
      };

   /// handle the scheduling of the approximation of the patches
   /// If a compiler which supports OpenMP, the implementation
   /// is parallel.
   template<typename T, int V>
      void Algo<T,V>::approximation(Processor<T,V>** processors,
            list<int>& listPatches, const int num_threads,
            const bool weightConstant,
            const int offsetCount) {
         for (int i = 0; i<num_threads;++i) {
            /// WARNING
            processors[i]->setWeightConstant(weightConstant);
         }
         if (offsetCount == 0) _dictionary->resetLearning();
         int numPatches = listPatches.size()+offsetCount;
         int numPatchesPerPercent = listPatches.size()/100;
         if (listPatches.size() > 100*numPatchesPerPercent) 
            numPatchesPerPercent++;
#ifdef _OPENMP
         list<int>::const_iterator patch = listPatches.begin();
         int count = offsetCount;
         int countT, patchT;
         countT = 0;
         patchT = 0;
         omp_set_num_threads(num_threads);
         omp_set_nested(false);
         omp_set_dynamic(false);
#pragma omp parallel for schedule(static,1) firstprivate(countT,patchT) shared(count,patch,processors) 
         for (int i = 0; i<num_threads; ++i) {
            while (patch != listPatches.end()) {
#pragma omp critical (assign) 
               {
#pragma omp flush(count,patch)
                  if (count >= numPatches) {
                     countT = -1;
                     patchT = -1;
                  } else {
                     countT = count;
                     patchT = *patch;
                     if (count % numPatchesPerPercent == 0) cout <<
                        "Percents processed : " <<
                           count/numPatchesPerPercent
                           << "% \r" << flush;
                     ++count;
                     ++patch;
                  }
               } // End of critical
               if (countT > -1) {
                  processors[i]->initialize(countT,patchT);
                  processors[i]->run();
               }
            }
         } // End of loop
#else
         Processor<T,V>* processor = processors[0];
         int count = 0;
         for ( list<int>::const_iterator patch = listPatches.begin();
               patch != listPatches.end();
               ++patch) {
            processor->initialize(count,*patch);
            processor->run();
            if (count % numPatchesPerPercent == 0) cout <<
               "Percents processed : " <<
                  count/numPatchesPerPercent
                  << "% \r" << flush;
            ++count;
         };
#endif // _OPENMP
         cout << "Percents processed : 100%" << endl;
      };

   namespace Sequence {
      template <typename T = double, int V = 1, int W = 3> class Algo {
         public:

            /// Constructor, original is optional
            /// input is a noisy image, if the original video 
            /// is provided, it will provides the PSNR values
            /// during the process
            Algo(const Video<T,V,W>* const input, 
                  const QString& fileDictionary, 
                  const Video<T,V,W>* const original = NULL,
                  const int overlapFactor = OVERLAP_FACTOR_VIDEO);

            /// Virtual Destructor
            virtual ~Algo();

            /// execute the algorithm virtual method.
            /// This should be implemented in the derived classes
            virtual void exec(const T gam) = 0; 
            /// print some information about the algorithm
            virtual void printInformation() const;
            /// print the PSNR between the original and the output
            void printPSNR() const;
            /// return the PSNR
            T inline PSNR() const;
            /// Accessor for the filename of the output.
            bool save(const QString& fileName) const;
            /// write the dictionary in an image
            void inline writeDictionaryToImage(const QString& file) const;

         protected:
            /// result of the algorithm
            Video<T,V,W>* _output;   
            /// input of the algorithm
            const Video<T,V,W>* const _input;    
            /// optional : original data.
            const Video<T,V,W>* const _original; 
            /// number of scales
            int _N;      
            /// dictionary
            Dictionary<T,V*W>* _dictionary;
            /// number of pixels between patches
            int _overlapFactor;
            /// size of the patches which define the overlapping level
            int _basePatch;
            /// print some statistics or not
            bool _stats;

         private:
            /// Copy constructor is forbidden
            explicit Algo(const Algo<T,V,W>& alg);
            /// operator= is forbidden
            Algo<T,V,W>& operator=(const Algo<T,V,W>& alg);
      };

      /// Constructor, original is optional
      /// input is a noisy image, if the original video 
      /// is provided, it will provides the PSNR values
      /// during the process
      template<typename T, int V, int W>
         Algo<T,V,W>::Algo(const Video<T,V,W>* const input, 
               const QString& fileDictionary, 
               const Video<T,V,W>* const original,
               const int overlapFactor) :
            _input(input), _original(original),
            _output(NULL), _stats(true), 
            _overlapFactor(overlapFactor) {
               _dictionary = new Dictionary<T,V*W>(fileDictionary);
               _N = _dictionary->N();
               if (_N > MAX_N) {
                  cerr << "Error : N too big" << endl;
                  exit(1);
               }
               _basePatch = static_cast<int>(static_cast<double>(
                        _dictionary->sizeEdgeBigPatches()));
            };

      /// print some information about the algorithm
      template<typename T, int V, int W>
         void Algo<T,V,W>::printInformation() const {
            cout << "Video Algorithm" << endl;
            cout << "The number of scales is " << _N << endl;
            cout << "The input image is : ";
            _input->printInformation();
            _dictionary->printInformation();
         };

      /// Destructor
      template<typename T, int V, int W>
         Algo<T,V,W>::~Algo() {
            delete(_dictionary);
            _dictionary = NULL;
            delete(_output);
            _output = NULL;
         };

      /// print the PSNR between the original and the output
      template<typename T, int V, int W>
         void Algo<T,V,W>::printPSNR() const {
            if (!_output || !_original) return;
            cout << "PSNR : " << PSNR() << endl; 
            if (_original) {
               cout << "PSNR initial : " << _input->PSNR(*_original) << endl; 
            }
         };

      /// return the PSNR
      template<typename T, int V, int W>
         T inline Algo<T,V,W>::PSNR() const {
            assert(_original != NULL && _output != NULL);
            return _output->PSNR(*_original);
         };

      /// Accessor for the filename of the output.
      template<typename T, int V, int W>
         bool Algo<T,V,W>::save(const QString& baseName) const {
            bool res = _output && _output->save(baseName+"/restored");
            if (_original)
               _original->save(baseName+"/orig");
            return res && _input && _input->save(baseName+"/input");
         };

      /// write the dictionary in an image
      template<typename T, int V, int W>
         void inline Algo<T,V,W>::writeDictionaryToImage(
               const QString& file) const {
            _dictionary->writeDictionaryToImage(file);
            _dictionary->writeDictionary(file+".dict");
         };
   };
};

#endif // ALGO_H
