/* demosaicing.h */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
*/

/*!
 * \file
 *                Multiscale KSVD Demosaicing
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File demosaicing.h
 * \brief Contains the class Demosaicing */

#ifndef DEMOSAICING_H
#define DEMOSAICING_H

using namespace std;

namespace KSVD {
   
   namespace Weighted {
      
      template <typename T> class Demosaicing : 
         public KSVD::Weighted::InpaintingDoubleLearning<T,3> {
            public:
               /// Constructor
               Demosaicing(const Image<T,3>* const input, 
                     const int J, const QString& fileDictionary,
                     const int maxL = DEFAULT_MAX_SPARSITY_INPAINTING,
                     const Image<T,3>* const original = NULL);

               /// print some information about the algorithm's parameters
               virtual void printInformation() const;
         };

      /// Constructor
      template <typename T>
         Demosaicing<T>::Demosaicing(
               const Image<T,3>* const input, 
               const int J, const QString& fileDictionary,
               const int maxL,
               const Image<T,3>* const original) :
            InpaintingDoubleLearning<T,3>(input,NULL,
                  J,fileDictionary,maxL,original) {
               delete(this->_map);
               this->_map = new Image<T,3>(this->_input->width(),
                     this->_input->height());
               this->_map->bayerPattern();
            };

      /// print some information about the algorithm's parameters.
      template <typename T>
         void Demosaicing<T>::printInformation() const {
            cout << "Executing Demosaicing Algorithm with "
               << "Multiscale KSVD" << endl;
            cout << "J = " << this->_J << endl;
            cout << "L = " << this->_maxL << endl;
            Algo<T,3>::printInformation();
         };
   };
   
   typedef KSVD::Weighted::Demosaicing<double> Demosaicing;

   namespace Sequence {
   
      // Video Demosaicing
      template <typename T, int W> class Demosaicing : 
         public KSVD::Sequence::InpaintingDoubleLearning<T,3,W> {
            public:
               /// Constructor
               Demosaicing(const Video<T,3,W>* const input, 
                     const int J, const QString& fileDictionary,
                     const int maxL = DEFAULT_MAX_SPARSITY_INPAINTING,
                     const Video<T,3,W>* const original = NULL,
                     const int Llearning = DEFAULT_LOW_SPARSITY,
                     const int overlapFactor = OVERLAP_FACTOR_VIDEO);

               /// print some information about the algorithm's parameters
               virtual void printInformation() const;
         };

      /// Constructor
      template<typename T, int W>
         Demosaicing<T,W>::Demosaicing(const Video<T,3,W>* const input, 
                     const int J, const QString& fileDictionary,
                     const int maxL,
                     const Video<T,3,W>* const original,
                     const int Llearning,
                     const int overlapFactor) :
            InpaintingDoubleLearning<T,3,W>(input,NULL,J,fileDictionary,
                  maxL,original,Llearning,overlapFactor) {
               delete(this->_map);
               this->_map = new Video<T,3,W>(this->_input->width(),
                     this->_input->height(), this->_input->nFrames());
               this->_map->bayerPattern();
            };

      /// print some information about the algorithm's parameters
      template<typename T, int W>
         void Demosaicing<T,W>::printInformation() const {
            cout << "Executing Demosaicing Algorithm with "
               << "Multiscale KSVD for Video" << endl;
            cout << "J = " << this->_J << endl;
            cout << "L = " << this->_maxL << endl;
            Algo<T,3,W>::printInformation();
         };
   };

   typedef KSVD::Sequence::Demosaicing<double,3> DemosaicingVideo;
   typedef KSVD::Sequence::Demosaicing<double,5> DemosaicingVideoHQ;

};

#endif

