/* dictionary.h */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
   */

/*!
 * \file
 *                Multiscale KSVD Denoising
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File dictionary.h
 * \brief Contains the class Dictionary */

#ifndef DICTIONARY_H
#define DICTIONARY_H

#include "image.h"
#include <fstream>
#ifdef SVDPOWER
#include "svdPower/svdPower.c"
#else 
#include "svdlibc/svdlib.h"
#endif

#define INDEX_SPARSE_MATRIX long int
/// max correlation authorized within the dictionary
#define MAX_SCAL_ATOMS 0.95
/// minimum usage authorized for each atom of the dictionary
#define MIN_ATOMS_USED 3

//static bool DEBUGVAR;
namespace KSVD {

   /*!
    * \brief
    * This class represents the dictionary used in the K-SVD.
    *
    * T is the type of data, V is the number of channels */
   template <typename T = double, int V = 1> class Dictionary {

      typedef TNT::Sparse_Matrix<T> Sparse_Matrix;
      typedef TNT::Sparse_Vector<T> Sparse_Vector;

      public:
      /// Constructor with file
      Dictionary(const QString& fileDictionary, 
            const bool addConstantAtoms = true);

      /// Constructor for external interfaces, 1-scale only
      Dictionary(T* dict,int n, int k);
      
      /// Constructor for external interfaces, multiscale
      Dictionary(T** dict,int* n, int* k, int N);

      /// Copy Constructor
      explicit Dictionary(const Dictionary<T,V>& Dictionary);
      /// Build a joint dictionary
      Dictionary(const Dictionary<T,V>& dict1, const Dictionary<T,V>& dict2);
      /// Destructor
      ~Dictionary();

      /// this function must be called before calling performLearning
      /// or any learning process
      void allocateLearning(int numPatches, const int Nt,
            bool fixedSparsity = false,
            bool allocateResidual = true);
      /// perform the learning step of the KSVD algorithm
      void performLearning(bool cleanDico = false,bool removeAtoms = false,
            const Image<T,V>* const image = NULL,
            const Image<T,V>* const beta = NULL);
      /// free some memory when the learning is finished.
      void terminateLearning();
      /// reset the coefficients to prepare for a new approximation
      inline void resetLearning();
      /// set the residual the patch indexed by "indexPatch"
      void inline updateResidual(int indexPatch, const VectR& residual);
      /// insert a coeff at the level k, for the atom index and a coeff value.
      void inline insertCoef(const int k,
            const int indexPatch, const int index, T value); 
      /// insert an indexPatch
      void inline insertIndexPatch(const int index, const int number);
      /// print some information
      void printInformation() const;
      /// print the sparsity
      void printSparsity() const;
      /// Change the structure of the patches
      void reverseDictStructure();
      /// write the dictionary to a file
      void writeDictionary(const QString& fileName) const;
      /// print dictionary in images files
      void writeDictionaryToImage(
            const QString& fileName) const;
      /// write the coefficients into a file
      void writeCoeffs(const QString& fileName) const;
      /// add the reconstruction of the patch to the image
      void inline addPatch(Image<T,V>& image,
            const int numPatch, const int indexPatch,
            const int sizeEdgePatches, const int N) const;
      /// switch to the RGB Metric
      void switchToRGBMetric(const T gamma);
      /// switch from the RGB Metric
      void switchFromRGBMetric(const T gamma);
      /// function to call when using external interfaces
      void inline setDataDictNull() {
         for (int i = 0; i<_N; ++i) {
            _dicts[i]->setArrayNull();
         }
      };

      /// Accessor
      int inline N() const { return _N; };
      /// Accessor 
      int inline n(int i) const {
         assert (i <_N && i >=0);
         return _n[i];
      };
      /// Accessor
      int inline K(int i) const {
         assert (i <_N && i >=0);
         return _K[i];
      };
      /// Accessor
      inline const Matrix& dicts(int i) const {
         assert (i <_N && i >=0);
         return *(_dicts[i]);
      };
      /// Accessor
      inline const Matrix& altDicts(int i) const {
         assert (i <_N && i >=0);
         return *(_altDicts[i]);
      };
      /// Accessor
      inline const Sparse_Matrix& coeffs(int i) const {
         assert (i <_N && i >=0);
         return *(_coeffs[i]);
      }
      /// Accessor
      inline T** invNorm() const {
         return _invNorm;
      };
      /// Accessor
      inline int sizeEdgeSmallPatches() const {
         return static_cast<int>(std::sqrt(static_cast<T>(_n[0]/V)));
      };
      /// Accessor
      inline int sizeEdgeBigPatches() const {
         return static_cast<int>(std::sqrt(static_cast<T>(_n[_N-1]/V)));
      };
      /// Accessor
      inline bool storeResidual() const {
         return _storeResidual;
      };
      /// Accessor
      inline bool addConstantAtoms() const {
         return _addConstantAtoms;
      };

      private:
      /// operator= is forbidden
      Dictionary<T,V>& operator=(const Dictionary<T,V>& dict);
      /// compute the reconstruction of one patch
      inline void computePatch(VectR& patch, const int numPatch,
            const int level, int offset = 0) const;
      /// clean the dictionary
      void cleanDictionary(Sparse_Matrix& coeffsTrans,
            const int level, bool removeAtoms = false);
      /// extract the right position in the residual for a patch value
      /// in the residual
      inline T& extractRes(const int& index, const int& level,
            const int& indexInPatch) const;
      /// compute the residual
      inline void extractPatch(const Image<T,V>& image,const int index,
            const int level,VectR& patch) const;

      /// store the residual or not
      bool _storeResidual;
      /// if the residual is not stored, then the index of the patches
      /// have to be stored
      int* _indexPatchesInImage;
      /// number of scales
      int _N;
      /// fixed sparsity or not
      bool _fixedSparsity;
      /// number of atoms for each level
      int* _K;
      /// size of atoms for each level
      int* _n;
      /// array, coefficients for each level of detail and each patch
      Sparse_Matrix** _coeffs;
      /// array, dictionaries for each level of detail 
      Matrix** _dicts;
      /// modified dictionaries for each level of detail 
      Matrix** _altDicts;
      /// inverse norm of the original modified dictionaries
      T** _invNorm;
      /// residuals
      Matrix* _residual;
      /// add constant atoms or not
      bool _addConstantAtoms;
   };

   /// Constructor with file
   template <typename T, int V>
      Dictionary<T,V>::Dictionary(const QString& fileDictionary,
            const bool addConstantAtoms) :
         _residual(NULL), _indexPatchesInImage(NULL),
         _addConstantAtoms(addConstantAtoms) {
            ifstream f(fileDictionary);
            if (!f) {
               cerr << "Failed to open dictionary file" << endl;
               exit(1);
            }
            cout << "Dictionary " << fileDictionary << " is being"
               << " opened" << endl;
            f >> _N;
            _K = new int[_N];
            _n = new int[_N];
            _dicts = new Matrix*[_N];
            _altDicts = new Matrix*[_N];
            _invNorm = new T*[_N];
            _residual = new Matrix();
            _coeffs = new Sparse_Matrix*[_N];
            for (int i = 0; i<_N; ++i) {
               _coeffs[i] = NULL;
            }

            for (int i = 0; i<_N; ++i) {
               f >> _n[i];
               f >> _K[i];
               int suppAtoms = _addConstantAtoms ? V : 0;
               _K[i] += suppAtoms;
               _dicts[i] = new Matrix(_n[i],_K[i]);
               _altDicts[i] = new Matrix(_n[i],_K[i]);
               _invNorm[i] = new T[_K[i]];
               int sizeEdge = static_cast<int>(std::sqrt(
                        static_cast<T>(_n[i]/V)));
               for (int j = 0; j<_K[i]-suppAtoms; ++j) {
                  VectR tmp(_n[i]);
                  for (int k = 0; k<_n[i]; ++k) {
                     f >> tmp(k);
                  }
                  CPPL::changeStructurePatch<T,V>(tmp,sizeEdge,i);
                  for (int k = 0; k<_n[i]; ++k) {
                     _dicts[i]->set(k,j,tmp(k));
                  }
               }
               if (_addConstantAtoms) {
                  // add the constant atom at the end of the dictionary
                  T tmp = 1.0/std::sqrt(static_cast<T>(_n[i]/V));
                  for (int k = 0; k<suppAtoms; ++k) {
                     for (int j = 0; j<_n[i]; ++j) {
                        _dicts[i]->set(j,_K[i]-suppAtoms+k,0.0);
                     }
                     for (int j = 0; j<_n[i]/V; ++j) {
                        _dicts[i]->set(V*j+k,_K[i]-suppAtoms+k,tmp);
                     }                  
                  }
               }
               for (int k = 0; k<_K[i]; ++k) {
                  _invNorm[i][k] = 1.0;
               }
            }
         };

   /// Constructor for external interfaces, 1-scale only
   template <typename T, int V>
      Dictionary<T,V>::Dictionary(T* data,int n, int k) :
         _residual(NULL), _indexPatchesInImage(NULL),
         _addConstantAtoms(false) {
            _N=1;
            _K = new int[_N];
            _n = new int[_N];
            _dicts = new Matrix*[_N];
            _altDicts = new Matrix*[_N];
            _invNorm = new T*[_N];
            _residual = new Matrix();
            _coeffs = new Sparse_Matrix*[_N];
            for (int i = 0; i<_N; ++i) {
               _coeffs[i] = NULL;
            }
            _n[0]=n;
            _K[0]=k;
            _dicts[0] = new Matrix(data,_n[0],_K[0]);
            _altDicts[0] = new Matrix(_n[0],_K[0]);
            _invNorm[0] = new T[_K[0]];
            for (int k = 0; k<_K[0]; ++k) {
               _invNorm[0][k] = 1.0;
            }
         };

   /// Constructor for external interfaces, multiscale
   template <typename T, int V>
      Dictionary<T,V>::Dictionary(T** data,int* n, int* k, int N) :
         _residual(NULL), _indexPatchesInImage(NULL),
         _N(N),
         _addConstantAtoms(false) {
            _K = new int[_N];
            _n = new int[_N];
            memcpy(_K,k,N*sizeof(int));
            memcpy(_n,n,N*sizeof(int));
            _dicts = new Matrix*[_N];
            _altDicts = new Matrix*[_N];
            _invNorm = new T*[_N];
            _residual = new Matrix();
            _coeffs = new Sparse_Matrix*[_N];
            for (int i = 0; i<_N; ++i) {
               int sizeEdge = static_cast<int>(std::sqrt(
                        static_cast<T>(_n[i]/V)));
               _coeffs[i] = NULL;
               _dicts[i] = new Matrix(data[i],_n[i],_K[i]);
               for (int j = 0; j<_K[i]; ++j) {
                  VectR tmp = _dicts[i]->colRef(j);
                  CPPL::changeStructurePatch<T,V>(tmp,sizeEdge,i);
                  tmp.setArrayNull();
               }
               _altDicts[i] = new Matrix(_n[i],_K[i]);
               _invNorm[i] = new T[_K[i]];
               for (int k = 0; k<_K[i]; ++k) {
                  _invNorm[i][k] = 1.0;
               }
            }
         };

   /// Copy constructor
   template <typename T, int V>
      Dictionary<T,V>::Dictionary(const Dictionary<T,V>& dict) :
         _N(dict._N), _storeResidual(dict._storeResidual),
         _indexPatchesInImage(NULL) {
            _addConstantAtoms=dict._addConstantAtoms;
            _K = new int[_N];
            _n = new int[_N];
            _dicts = new Matrix*[_N];
            _altDicts = new Matrix*[_N];
            _invNorm = new T*[_N];
            _residual = new Matrix();
            _coeffs = new Sparse_Matrix*[_N];
            for (int i = 0; i<_N; ++i) {
               _coeffs[i] = NULL;
               _n[i] = dict._n[i];
               _K[i] = dict._K[i];
               _dicts[i] = new Matrix(*(dict._dicts[i]));
               _altDicts[i] = new Matrix(*(dict._altDicts[i]));
               _invNorm[i] = new T[_K[i]];
               for (int k = 0; k<_K[i]; ++k) {
                  _invNorm[i][k] = dict._invNorm[i][k];
               }
            }
         };

   /// Build a joint dictionary
   template <typename T, int V>
      Dictionary<T,V>::Dictionary(const Dictionary<T,V>& dict1,
            const Dictionary<T,V>& dict2) :
         _N(dict1._N), _storeResidual(dict1._storeResidual),
         _indexPatchesInImage(NULL) {
            _addConstantAtoms=dict1._addConstantAtoms ||
               dict2._addConstantAtoms;
            _K = new int[_N];
            _n = new int[_N];
            _dicts = new Matrix*[_N];
            _altDicts = new Matrix*[_N];
            _residual = new Matrix();
            _coeffs = new Sparse_Matrix*[_N];
            _invNorm = new T*[_N];
            for (int i = 0; i<_N; ++i) {
               _coeffs[i] = NULL;
               _n[i] = dict1._n[i];
               /// remove the constant atoms
               int suppAtomsDict1 = dict1._addConstantAtoms ? V : 0;
               int suppAtomsDict2 = dict2._addConstantAtoms ? V : 0;
               int suppAtoms = _addConstantAtoms ? V : 0;
               _K[i] = dict1._K[i] - suppAtomsDict1 + dict2._K[i] - suppAtomsDict2 + suppAtoms;
               _dicts[i] = new Matrix(_n[i],_K[i]);
               _altDicts[i] = new Matrix(_n[i],_K[i]);
               _invNorm[i] = new T[_K[i]];
               for (int j = 0; j<dict1._K[i]-suppAtomsDict1; ++j) {
                  VectR col = dict1._dicts[i]->colRef(j);
                  _dicts[i]->setColumn(col,j);
                  col.setArrayNull();
                  VectR col2 = dict1._altDicts[i]->colRef(j);
                  _altDicts[i]->setColumn(col2,j);
                  col2.setArrayNull();
                  _invNorm[i][j] = dict1._invNorm[i][j];
               }
               for (int j = 0; j<dict2._K[i]-suppAtomsDict2; ++j) {
                  VectR col = dict2._dicts[i]->colRef(j);
                  _dicts[i]->setColumn(col,j+dict1._K[i]-suppAtomsDict1);
                  col.setArrayNull();
                  VectR col2 = dict2._altDicts[i]->colRef(j);
                  _altDicts[i]->setColumn(col2,j+dict1._K[i]-suppAtomsDict1);
                  col2.setArrayNull();
                  _invNorm[i][j+dict1._K[i]-suppAtomsDict1] = dict2._invNorm[i][j];
               }
               if (_addConstantAtoms) {
                  // add the constant atom at the end of the dictionary
                  T tmp = 1.0/std::sqrt(static_cast<T>(_n[i]/V));
                  for (int k = 0; k<suppAtoms; ++k) {
                     for (int j = 0; j<_n[i]; ++j) {
                        _dicts[i]->set(j,_K[i]-suppAtoms+k,0.0);
                     }
                     for (int j = 0; j<_n[i]/V; ++j) {
                        _dicts[i]->set(V*j+k,_K[i]-suppAtoms+k,tmp);
                     }                  
                  }
                  for (int k = 0; k<suppAtoms; ++k) {
                     _invNorm[i][_K[i]-k-1] = 1.0;
                  }
               }
            }
         };

   /// Destructor
   template <typename T, int V>
      Dictionary<T,V>::~Dictionary() {
         delete[](_K);
         delete[](_n);
         for (int i = 0; i<_N; ++i) {
            delete(_coeffs[i]);
            _coeffs[i] = NULL;
         }
         delete[](_coeffs);
         for (int i = 0; i<_N; ++i) {
            delete(_dicts[i]);
            _dicts[i] = NULL;
            delete(_altDicts[i]);
            _altDicts[i] = NULL;
            delete[](_invNorm[i]);
            _invNorm[i] = NULL;
         }
         delete[](_dicts);
         delete[](_altDicts);
         delete[](_invNorm);
         delete(_residual);
         delete[](_indexPatchesInImage);
      };

   /// this function must be called before calling performLearning
   /// or any learning process
   template <typename T, int V>
      void Dictionary<T,V>::allocateLearning(int numPatches,
            const int Nt, bool fixedSparsity,
            bool allocateResidual) {
         _storeResidual = allocateResidual;
         _fixedSparsity = fixedSparsity;
         numPatches *= fourPow[Nt-_N];
         _indexPatchesInImage = new int[numPatches];
         if (_storeResidual) {
            _residual->resize(_n[_N-1],numPatches);
         }
         for (int i = _N-1; i>=0; --i) {
            _coeffs[i] = new Sparse_Matrix(numPatches,_K[i]);
            numPatches *= 4;
         }
      };

   /// perform the learning step of the KSVD algorithm
   template <typename T, int V>
      void Dictionary<T,V>::performLearning(bool cleanDico,
            bool removeAtoms,
            const Image<T,V>* const image,
            const Image<T,V>* const beta) {
         assert(_storeResidual || image != NULL);
         bool weighted = beta != NULL;
         for (int level = 0; level<_N; ++level) {
            int numPatches = _coeffs[level]->num_rows();
            Sparse_Matrix coeffsTrans(_K[level],numPatches);
            // transpose the coeff matrix
            for (int i = 0; i<numPatches; ++i) {
               const Sparse_Vector& alpha = _coeffs[level]->row(i);
               for (typename Sparse_Vector::const_iterator
                     it = alpha.begin();
                     it != alpha.end();
                     ++it) {
                  coeffsTrans.insert(it->value(),it->index(),i);
               }
            }
            coeffsTrans.close();

            // Do not update the constant atoms
            int suppAtoms = _addConstantAtoms ? V : 0;
            for (int i = 0; i<_K[level]-suppAtoms; ++i) {
               int num_coeffs = coeffsTrans.row(i).num_nonzeros();
               if (num_coeffs != 0) {
                  int sizeMatrix = num_coeffs*_n[level];
#ifdef SVDPOWER
                  T* M = new T[sizeMatrix];
                  T* u = new T[_n[level]];
                  for (int j = 0; j<_n[level]; ++j) u[j] = 0.0;
                  T* v = new T[num_coeffs];
                  for (int j = 0; j<num_coeffs; ++j) v[j] = 0.0;
                  T* d = new T[_n[level]];
                  for (int j = 0; j<_n[level]; ++j) d[j] = (*(_dicts[level]))(j,i);

                  int l = 0;
                  for (typename Sparse_Vector::const_iterator
                        it = coeffsTrans.row(i).begin();
                        it != coeffsTrans.row(i).end();
                        ++it) {
                     T* res = &(extractRes(it->index(),level,0));
                     for (int j = 0; j<_n[level]; ++j) {
                        M[j+l*_n[level]] = *(res+j) +
                           it->value()* d[j];
                     }
                     ++l;
                  }
                  svdPower(M,_n[level],num_coeffs,d,u,v);
                  free(M);
                  free(d);
                  l = 0;
                  for (typename Sparse_Vector::iterator
                        it = coeffsTrans.row(i).begin();
                        it != coeffsTrans.row(i).end();
                        ++it) {
                     if (_storeResidual) {
                        if (weighted) {
                           VectR weightPatch(_n[level]);
                           extractPatch(*beta,it->index(),level,
                                 weightPatch);
                           T* res = &(extractRes(it->index(),level,0));
                           for (int j = 0; j<_n[level]; ++j) {
                              *(res+j) -= weightPatch(j)*(
                                    u[j]*v[l] - it->value()*
                                    (*(_dicts[level]))(j,i));
                           }
                        } else {
                           T* res = &(extractRes(it->index(),level,0));
                           for (int j = 0; j<_n[level]; ++j) {
                              *(res+j) -=
                                 u[j]*v[l] - it->value()*
                                 (*(_dicts[level]))(j,i);
                           }
                        }
                     }
                     it->value() = v[l];
                     ++l;
                  }
                  for (int j = 0; j<_n[level]; ++j) {
                     (*(_dicts[level]))(j,i) = u[j];
                  }

                  free(u);
                  free(v);
#else
                  T* values = new T[sizeMatrix];
                  INDEX_SPARSE_MATRIX* irow = 
                     new INDEX_SPARSE_MATRIX[sizeMatrix];
                  INDEX_SPARSE_MATRIX* pcol = 
                     new INDEX_SPARSE_MATRIX[num_coeffs+1];
                  for (int j = 0; j<=num_coeffs; ++j) {
                     pcol[j] = j*_n[level];
                  }
                  int l = 0;
                  for (typename Sparse_Vector::const_iterator
                        it = coeffsTrans.row(i).begin();
                        it != coeffsTrans.row(i).end();
                        ++it) {
                     if (_storeResidual) {
                        T* res = &(extractRes(it->index(),level,0));
                        for (int j = 0; j<_n[level]; ++j) {
                           irow[j+l*_n[level]] = j;
                        }
                        for (int j = 0; j<_n[level]; ++j) {
                           values[j+l*_n[level]] = *(res+j) +
                              it->value()* (*(_dicts[level]))(j,i);
                        }
                     } else {
                        // TODO
                        /*   VectR patch(_n[level]);
                             extractRes(image,it->index(),level,patch);
                             for (int j = 0; j<_n[level]; ++j) {
                             irow[j+l*_n[level]] = j;
                             values[j+l*_n[level]] = patch[j] +
                             it->value()* (*(_dicts[level]))(j,i);
                             }*/
                     }
                     ++l;
                  }
                  SMat A = (SMat)malloc(sizeof(struct smat));
                  A->rows = _n[level];
                  A->cols = num_coeffs;
                  A->vals = sizeMatrix;
                  A->pointr = pcol;
                  A->rowind = irow;
                  A->value = values;
                  SVDRec svd = svdLAS2A(A,1);
                  free(A);
                  l = 0;
                  if (svd->S[0] > EPSILON) {
                     for (typename Sparse_Vector::iterator
                           it = coeffsTrans.row(i).begin();
                           it != coeffsTrans.row(i).end();
                           ++it) {
                        if (_storeResidual) {
                           if (weighted) {
                              VectR weightPatch(_n[level]);
                              extractPatch(*beta,it->index(),level,
                                    weightPatch);
                              T* res = &(extractRes(it->index(),level,0));
                              for (int j = 0; j<_n[level]; ++j) {
                                 *(res+j) -= weightPatch(j)*(
                                       svd->S[0]*svd->Ut->value[0][j]*
                                       svd->Vt->value[0][l] - it->value()*
                                       (*(_dicts[level]))(j,i));
                              }
                           } else {
                              T* res = &(extractRes(it->index(),level,0));
                              for (int j = 0; j<_n[level]; ++j) {
                                 *(res+j) -=
                                    svd->S[0]*svd->Ut->value[0][j]*
                                    svd->Vt->value[0][l] - it->value()*
                                    (*(_dicts[level]))(j,i);
                              }
                           }
                        }
                        it->value() = svd->S[0]*svd->Vt->value[0][l];
                        ++l;
                     }
                     for (int j = 0; j<_n[level]; ++j) {
                        (*(_dicts[level]))(j,i) = svd->Ut->value[0][j];
                     }
                  }
                  svdFreeSVDRec(svd);
                  delete[](values);
                  delete[](irow);
                  delete[](pcol);
#endif
               } 
            }
            delete(_coeffs[level]);
            _coeffs[level] = new Sparse_Matrix(numPatches,_K[level]);
            // transpose the coeff matrix
            for (int i = 0; i<_K[level]; ++i) {
               const Sparse_Vector& alpha = coeffsTrans.row(i);
               for (typename Sparse_Vector::const_iterator
                     it = alpha.begin();
                     it != alpha.end();
                     ++it) {
                  _coeffs[level]->insert(it->value(),it->index(),i);
               }
            }
            if (cleanDico) cleanDictionary(coeffsTrans,level,removeAtoms);
         }
      };

   /// free some memory when the learning is finished.
   template <typename T, int V>
      void Dictionary<T,V>::terminateLearning() {
         _residual->clear();
         for (int i = 0; i<_N; ++i) {
            delete(_coeffs[i]);
            _coeffs[i] = NULL;
         }
         delete[](_indexPatchesInImage);
         _indexPatchesInImage = NULL;
      };

   /// reset the dictionary for a new approximation
   template <typename T, int V>
      inline void Dictionary<T,V>::resetLearning() {
         /// reset the data for the learning
         for (int i = _N-1; i>=0; --i) {
            int numPatches = _coeffs[i]->num_rows();
            delete(_coeffs[i]);
            _coeffs[i] = new Sparse_Matrix(numPatches,_K[i]);
         }
      };

   /// set the residual the patch indexed by "indexPatch"
   template <typename T, int V>
      void inline Dictionary<T,V>::updateResidual(int indexPatch,
            const VectR& residual) {
         assert(_storeResidual);
         assert(check(residual));
#pragma omp critical
         {
            _residual->setColumn(residual,indexPatch);
         }
      };

   /// insert a coeff at the level k, for the atom index and a coeff value.
   template <typename T, int V>
      void inline Dictionary<T,V>::insertCoef(const int k,
            const int indexPatch,
            const int index, T value) {
         //  if (DEBUGVAR) {
//                           cerr << k << " " << indexPatch << " " << index << " "
 //                             << value << endl;
         //  }
         assert(!isnan(value));
#pragma omp critical
         {
            _coeffs[k]->insert(value,indexPatch,index);
         }
      };

   /// insert an index Patch
   template <typename T, int V>
      void inline Dictionary<T,V>::insertIndexPatch(const int index,
            const int number) {
         _indexPatchesInImage[number] = index;
      };

   /// print so/e information
   template <typename T, int V>
      void Dictionary<T,V>::printInformation() const {
         cout << "Number of dictionary scales : " << _N << endl;
         int n1 = sizeEdgeBigPatches();
         int n2 = sizeEdgeSmallPatches();
         cout << "Size of the patches at the top level : "
            << n1 << " x " << n1 << endl;
         cout << "Size of the patches at the low level : "
            << n2 << " x " << n2 << endl;
         cout << "Number of atoms " << _K[0] << endl;
      };

   /// print the sparsity
   template <typename T, int V>
      void Dictionary<T,V>::printSparsity() const {
         assert(_coeffs[0] != NULL);
         double sparsity = 0.0;
         for (int i = 0; i<_N; ++i) {
            double sparsityLevel = _coeffs[i]->num_nonzeros();
            sparsity += sparsityLevel;
            cout << "sparsity at level " << i
               << " : " << sparsityLevel/_coeffs[_N-1]->num_rows()
               << endl;
         }
         sparsity /= _coeffs[_N-1]->num_rows();
         cout << "total sparsity : " << sparsity << endl;

      };

   /// change the structure of the patches
   template <typename T, int V>
      void Dictionary<T,V>::reverseDictStructure() {
         for (int i = 0; i<_N; ++i) {
            for (int j = 0; j<_K[i]; ++j) {
               VectR tmp = _dicts[i]->colRef(j);
               CPPL::reverseStructurePatch<T,V>(tmp,i);
               tmp.setArrayNull();
            }
         }
      };

   /// write the dictionary to a file
   template <typename T, int V>
      void Dictionary<T,V>::writeDictionary(
            const QString& fileName) const {
         std::ofstream f;
         f.precision(12);
         f.flags(std::ios_base::scientific);
         f.open(fileName, ofstream::trunc);
         cout << "Dictionary written in " << fileName << endl;
         f << _N << endl;
         int suppAtoms = _addConstantAtoms ? V : 0;

         // Do not write the constant atoms
         for (int i = 0; i<_N; ++i) {
            f << _n[i] << " " << _K[i]-suppAtoms << endl;
            for (int j = 0; j<_K[i]-suppAtoms; ++j) {
               VectR w(_dicts[i]->col(j));
               CPPL::reverseStructurePatch<T,V>(w,i);
               for (int k = 0; k<_n[i]; ++k) {
                  f << w(k) << " ";
               }
               f << endl;
            }
         }
         f.close();
      };

   /// write the coefficients to a file
   template <typename T, int V>
      void Dictionary<T,V>::writeCoeffs(const QString& fileName) const {
         std::ofstream f;
         f.precision(12);
         f.flags(std::ios_base::scientific);
         f.open(fileName, ofstream::trunc);
         f << _N << endl;

         for (int i = 0; i<_N; ++i) {
            f << _coeffs[i]->num_cols() << " ";
            f << _coeffs[i]->num_rows() << " ";
            f << _coeffs[i]->num_nonzeros() << endl;
            for (int j = 0; j<_coeffs[i]->num_rows(); ++j) {
               const Sparse_Vector& alpha = _coeffs[i]->row(j);
               for (typename Sparse_Vector::const_iterator
                     it = alpha.begin();
                     it != alpha.end();
                     ++it) {
                  f << it->index() << " ";
                  f << j << " ";
                  f << it->value() << endl;
               }
            }
         }
         f.close();
      };

   /// print dictionary in images files
   template <typename T, int V>
      void Dictionary<T,V>::writeDictionaryToImage(
            const QString& fileName) const {

         for (int i = 0; i<_N; ++i) {
            QString endfix = "Dict";
            endfix.append(QString::number(i));
            QString file = fileName;
            file.insert(file.length()-4,endfix);
            int sizeEdge = static_cast<int>(std::sqrt(
                     static_cast<double>(_n[i]/V)));
            int numberRepresented = static_cast<int>(floor(std::sqrt(
                        static_cast<double>(_K[i]))));
            int sizeImage = numberRepresented*(sizeEdge+1)+1;
            QImage dict(sizeImage,sizeImage,32);
            QImage** dicts = NULL;
            if (V == 9) {
               dicts = new QImage*[3];
               for (int j = 0; j<3; ++j) 
                  dicts[j] = new QImage(sizeImage,sizeImage,32);
            } else if (V != 1 && V != 3) {
               dicts = new QImage*[V];
               for (int j = 0; j<V; ++j) 
                  dicts[j] = new QImage(sizeImage,sizeImage,32);
            }
            uchar* pixel = dict.bits();
            for (int j = 0; j<dict.numBytes(); ++j) {
               *pixel = 0;
               ++pixel;
            }
            int row = 1;
            int column = 1;
            int numAtom = 0;
            int indexAtom = 0;
            for (int j = 0; j<numberRepresented; ++j) {
               for (int k = 0; k<numberRepresented; ++k) {
                  row = 1+k*(sizeEdge+1);
                  column = 1+j*(sizeEdge+1);
                  VectR patch(_dicts[i]->col(numAtom));
                  CPPL::reverseStructurePatch<T,V>(patch,i);
                  CPPL::normalizePatchDisplay(patch);
                  indexAtom = 0;
                  int lastColumn = column;
                  for (int l=0; l<sizeEdge; ++l) {
                     column = lastColumn;
                     for (int m=0; m<sizeEdge; ++m) {
                        if (V==1) {
                           int val = static_cast<int>(patch(indexAtom));
                           dict.setPixel(column,row,
                                 qRgb(val,val,val));
                           indexAtom++;
                        } else if (V==3) {
                           int r = static_cast<int>(patch(indexAtom));
                           ++indexAtom;
                           int g = static_cast<int>(patch(indexAtom));
                           ++indexAtom;
                           int b = static_cast<int>(patch(indexAtom));
                           ++indexAtom;
                           dict.setPixel(column,row,
                                 qRgb(r,g,b));
                        } else if (V == 9) {
                           for (int o = 0; o<3; ++o) {
                              int r = static_cast<int>(patch(indexAtom));
                              ++indexAtom;
                              int g = static_cast<int>(patch(indexAtom));
                              ++indexAtom;
                              int b = static_cast<int>(patch(indexAtom));
                              ++indexAtom;
                              dicts[o]->setPixel(column,row,
                                    qRgb(r,g,b));
                           }
                        } else {
                           for (int o = 0; o<V; ++o) {
                              int val = static_cast<int>(patch(indexAtom));
                              dict.setPixel(column,row,
                                    qRgb(val,val,val));
                              ++indexAtom;
                           }
                        }
                        ++column;
                     }
                     ++row;
                  }
                  ++numAtom;
               }
            }
            if (V == 9) {
               for (int j = 0; j<3; ++j) {
                  QString file2(file);
                  file2.insert(file.length()-4,j);
                  dicts[j]->save(file2,"PNG");
                  delete(dicts[j]);
               }
               delete[](dicts);
            } else if (V != 1 && V != 3) {
               for (int j = 0; j<V; ++j) {
                  QString file2(file);
                  file2.insert(file.length()-4,j);
                  dicts[j]->save(file2,"PNG");
                  delete(dicts[j]);
               }
               delete[](dicts);
            } else {
               dict.save(file,"PNG");
               cout << "Dictionary " << i << " saved in " << file << endl;
            }
         }

      };

   /// add the reconstruction of the patch to the image
   template <typename T, int V>
      void inline Dictionary<T,V>::addPatch(Image<T,V>& image,
            const int numPatch, const int indexPatch,
            const int sizeEdgePatches, const int Nt) const {
         assert(Nt >= _N);
         const int n = twoPow[Nt-_N];
         const int n2 = n*n;
         const int newSizeEdge = sizeEdgePatches/n;
         // Warning, the quadtree structure is not respected here
         for (int i = 0; i<n; ++i) {
            for (int j = 0; j<n; ++j) {
               VectR* patch = new VectR(_n[_N-1]);
               patch->zero();
               computePatch(*patch,indexPatch*n2+i*n+j,_N-1);
               image.addPatch(*patch,numPatch+(i*image.width()+j)*
                     newSizeEdge,newSizeEdge,_N);
               delete(patch);
            }
         }
      };

   /// switch to the RGB Metric
   template <typename T, int V>
      void Dictionary<T,V>::switchToRGBMetric(const T gamma) {
         for (int i = 0; i<_N; ++i) {
            _altDicts[i]->copy(*(_dicts[i]));
            switchDictToRGBMetric<T,V>(*(_altDicts[i]),
                  _n[0],fourPow[i],gamma);
            normalizeDict(*(_altDicts[i]), _invNorm[i]);
         }
      };

   /// switch from the RGB Metric
   template <typename T, int V>
      void Dictionary<T,V>::switchFromRGBMetric(const T gamma) {
         for (int i = 0; i<_N; ++i) {
            _altDicts[i]->copy(*(_dicts[i]));
            switchDictToRGBMetric<T,V>(*(_altDicts[i]),
                  _n[0],fourPow[i],-gamma/(1.0+gamma));
            normalizeDict(*(_altDicts[i]), _invNorm[i]);
            for (int j = 0; j<_n[i]; ++j) {
               _invNorm[i][j] = 1.0;
            }
         }
      };

   /// compute the reconstruction of one patch
   template <typename T, int V>
      inline void Dictionary<T,V>::computePatch(VectR& patch, 
            const int numPatch, const int level,
            int offset) const {
         assert(level >= 0);
         CPPL::addMultSparse(*(_dicts[level]),
               _coeffs[level]->row(numPatch),patch);
         if (level == 0) return;
         const int newL = _dicts[level-1]->m;
         VectR* quad;
         for (int i = 0; i<4; ++i) {
            quad = new VectR(newL,patch,i);
            computePatch(*quad,4*numPatch+i,level-1);
            quad->setArrayNull();
            delete(quad);
         }
      };

   /// clean the dictionary
   /// WARNING : no reconstruction is possible after calling this function.
   template <typename T, int V>
      void Dictionary<T,V>::cleanDictionary(
            Sparse_Matrix& coeffsTrans, const int level,
            const bool removeAtoms) {
         int suppAtoms = _addConstantAtoms ? V : 0;
         if (removeAtoms) {
            std::vector<bool> toRemove(_K[level]-suppAtoms);
            int numToRemove = 0;
            for (int i = 0; i<_K[level]-suppAtoms; ++i) {
               toRemove[i] = (coeffsTrans.row(i).num_nonzeros()
                     < MIN_ATOMS_USED);
               if (toRemove[i]) numToRemove++;
            }
            int firstEmptyCol = -1;
            int colToMove = -1;
            _K[level] -= numToRemove;
            Matrix* newDict = new Matrix(_n[level],_K[level]);
            int index = 0;
            for (int i = 0; i<_K[level]+numToRemove; ++i) {
               if (i >= _K[level]+numToRemove-suppAtoms || !toRemove[i]) {
                  newDict->setColumn(_dicts[level]->col(i),index);
                  ++index;
               }
            };
            delete(_dicts[level]);
            _dicts[level] = newDict;
            cout << numToRemove << " atoms removed" << endl;
         } else {
            // TODO : prepare for !storeResidual
            // optimize this function which is too slow
            int numPatches = _coeffs[level]->num_rows();
            VectR norms(numPatches);
            int numPerColumn = fourPow[_N-1-level];
            int numColumn = numPatches/numPerColumn;
            int count = 0;
            int n = _n[level];
            if (_fixedSparsity) {
               for (int i = 0; i<numColumn; ++i) {
                  for (int j = 0; j<numPerColumn; ++j) {
                     norms(count) = dnrm2_(n,_residual->array+
                           _n[_N-1]*i+j*_n[level],1);
                     ++count;
                  }
               }
            } else {
               for (count = 0; count<numPatches; ++count) {
                  norms(count) = _coeffs[level]->row(count).num_nonzeros();
                  if (level >= 1) {
                     for (int i = 0;i<4; ++i) {
                        norms(count) += _coeffs[level-1]->
                           row(count*4+i).num_nonzeros();
                     }
                  }
               }
            }
            count = 0;
            /// do not remove the constant atoms
            for (int i = 0; i<_K[level]-suppAtoms; ++i) {
               int num_coeffs = coeffsTrans.row(i).num_nonzeros();
               if (num_coeffs < MIN_ATOMS_USED) {
                  //cerr << count << " " << norms;
                  //KSVD::stop();
                  ++count;
                  bool ok = false;
                  int index, column, shift, l;
                  T invNormCol, nor;
                  VectR col(_n[_N-1]);
                  VectR* atom = new VectR();
                  do {
                     col.zero();
                     index = idamax(norms);
                     column = index/numPerColumn;
                     shift = index-column*numPerColumn;
                     // TODO optimize this part
                     computePatch(col,column,_N-1);
                     col += _residual->col(column);
                     l = col.l/numPerColumn;
                     atom->setArrayNull();
                     delete(atom);
                     atom = new VectR(l,col,shift);
                     invNormCol = 1.0/nrm2(*atom);
                     VectR scal = (*atom) * (*(_dicts[level]));
                     ok = (fabs(damax(scal))*invNormCol) < MAX_SCAL_ATOMS;
                     T test =(fabs(damax(scal))*invNormCol);
                     nor = norms(index);
                     // if (ok) cerr << nor << " "
                     //    << fabs(damax(scal))*invNormCol << endl;
                     norms(index) = 0;
                  } while (!ok && nor != 0);
                  //           cerr << i << endl;
                  _dicts[level]->setColumn(*atom,i);
                  _dicts[level]->scalColumn(invNormCol,i);
                  atom->setArrayNull();
                  delete(atom);
               }
            }
            cout << count << " atoms replaced" << endl;
         }
      };

   /// extract the right position in the residual for a patch value
   /// in the residual
   template <typename T, int V>
      inline T& Dictionary<T,V>::extractRes(const int& index, 
            const int& level, const int& indexInPatch) const {
         int indexInMatrix = index/fourPow[_N-1-level];
         int shiftInImage = index - indexInMatrix*fourPow[_N-1-level];
         return (*_residual)(shiftInImage*_n[level]+
               indexInPatch,indexInMatrix);
      };

   /// compute the residual
   template <typename T, int V>
      inline void Dictionary<T,V>::extractPatch(const Image<T,V>& image,
            const int index,
            const int level,VectR& patch) const {
         int indexInMatrix = index/fourPow[_N-1-level];
         int indexInPatch = index - fourPow[_N-1-level]*indexInMatrix;
         const int& indexBigPatch = _indexPatchesInImage[indexInMatrix];
         const int& n = patch.l;
         int sizeEdge = static_cast<int>(sqrt(static_cast<T>(n/V)));
         int x = indexInPatch/twoPow[_N-1-level];
         int y = indexInPatch-x*twoPow[_N-1-level];
         int ind = indexBigPatch+(x*image.width()+y)*sizeEdge;
         image.setPatch(patch,ind,sizeEdge,level+1);
      };

};

#endif // DICTIONARY_H
