/* image.h */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
*/

/*!
 * \file
 *                Multiscale KSVD Denoising
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File image.h
 * \brief Contains the class Image */

#ifndef IMAGE_H
#define IMAGE_H

#include <stdlib.h>
#include <iostream>
#include <qstring.h> 
#include "misc.h"

#define EPSILON_IMAGE 0.5

namespace KSVD {

   /*!
    * \brief "image" class used for the K-SVD.
    *
    * The elements are of type T, V is the number of channels */
   template <typename T = double, int V = 1> class Image {
      public:
         /// Empty constructor
         Image();
         /// Constructor : the data will be copied
         Image(const T* const _data, const int w, const int h);
         /// Constructor : no input data
         Image(const int w, const int h);
         /// explicit Constructor : input from a file
         explicit Image(const QString& fileName);
         /// explicit Constructor : input from a QImage
         explicit Image(const QImage& fileName);
         /// Copy Constructor
         explicit Image(const Image<T,V>& image);

         /// Destructor
         ~Image() { delete[](_dataImage); };

         /// Operator=
         inline void operator=(const Image<T,V>& image);         
         /// Operator/
         inline void operator/=(const Image<T,V>& image);         
         /// Operator*
         inline void operator*=(const Image<T,V>& image);         
         /// Operator+
         inline void operator+=(const Image<T,V>& image);         
         /// Operator-
         inline void operator-=(const Image<T,V>& image);         
         /// Operator= for a scalar
         inline void operator=(const T lambda);
         /// Operator*
         inline void operator*=(const T lambda);
         /// Operator/
         inline void operator/=(const T lambda);
         /// Operator-
         inline void operator-=(const T lambda);
         /// Operator+
         inline void operator+=(const T lambda);
         /// Operator []
         inline T& operator[](const int index);
         /// Operator []
         inline T operator[](const int index) const;
         /// operator()
         inline T& operator()(const int i, const int j) const;
         /// operator()
         inline T& operator()(const int i, 
               const int j, const int k) const;
         /// return the max element of the image
         inline T max() const;
         /// return the min element of the image
         inline T min() const;
         /// replace each value by 1/value
         inline void inv();
         /// uncenter the data
         inline void unCenter();
         /// erase the image
         inline void clear();
         /// clip the data to the right range
         void clipData();
         /// save the image in a file
         bool saveFile(const QString& fileName) const;
         /// print some information about the image
         void printInformation() const;
         /// add a white gaussian noise to the image
         void addNoiseGaussian(const T sigma);
         /// add a white non-uniform gaussian noise to the image
         void addNoiseGaussian(const Image<T,V>& mapNoise);
         /// add a white non-uniform gaussian noise to the image
         void addNoiseGaussian(const QString& fileMapNoise);
         /// apply a map for future inpainting problem
         void applyMap(const Image<T,V>& map);
         /// put a Bayer pattern inside the image
         void bayerPattern();
         /// mozaic the image
         void mozaic();
         /// copy the patch number num in the vector patch
         void inline setPatch(VectR& patch, const int num,
               const int sizeEdgePatches, const int level = 1,
               const int offset = 0) const;
         /// increment the value of one area, which corresponds to
         /// the patch i, with a n x n patch
         inline void incrPatch(const int i, const int n);
         /// add the value of the patch in the image
         inline void addPatch(const VectR& patch,
               const int num, const int sizeEdgePatches,
               const int level = 1, const int offset = 0);
         /// compute the PSNR with the given image
         T PSNR(const Image<T,V>& input) const;
         /// return a new patch index for lower scales
         template <bool Top, bool Left>
            inline int quad(const int numPatch,
                  const int newSizeEdge) const;
         /// binarize
         void binarize();
         /// reduce the size without linear filtering
         void reduceScale(const int scale = 2);
         /// convolution by a small kernel (no FFT)
         void convolution(const Matrix& kernel);
         /// convolution by a small separable kernel (no FFT)
         void convolutionSeparable(const VectR& a, const VectR& b);
         /// blur using a kernel in a file
         void blur(const QString& fileKernel);

         /// Accessor for width
         inline int width() const { return _width; };
         /// Accessor for height
         inline int height() const { return _height; };
         /// Accessor for numData
         inline int numData() const { return _numData; };

       /*  /// for DEBUG
         Image<T,V>* quad(int i);
         /// for DEBUG
         void setQuad(const Image<T,V>& im, int i);*/

         /// extract region of interest
         Image<T,V>* regionOfInterest(int i, int j, int sizeX, int sizeY) const;
         /// add im to a region of interest
         void addToRegionOfInterest(const Image<T,V>& im, int i, int j);
         /// extract channel
         Image<T,1>* channel(int k) const;
         /// set channel
         void setChannel(const Image<T,1>& im, int k);

      protected:
         /// load a QImage
         void loadQImage(const QImage& image);
         /// clip a value from 0 to 255;
         static inline T clip(const T in) { 
            return MIN(MAX(in,MIN_VALUE-MEAN_VALUE),
                  MAX_VALUE-MEAN_VALUE);
         };

         /// _data of the image
         T* _dataImage;
         /// width
         int _width;
         /// height
         int _height;          
         /// size of one row
         int _sizeRow;
         /// size of _dataImage
         int _numData;      
   };

   /// grayscale image
   typedef Image<double,1> ImageGray;
   /// color image
   typedef Image<double,3> ImageRGB;

   /// Empty constructor
   template <typename T, int V>
      Image<T,V>::Image() : _dataImage(NULL),
      _width(0),_height(0),_numData(0),_sizeRow(0) { };

   /// Constructor : the data will be copied
   template <typename T, int V>
      Image<T,V>::Image(const T* const _data, const int w, 
            const int h) : 
         _width(w), _height(h), _numData(h*w*V), _sizeRow(V*w) {
            _dataImage = new T[_numData]; 
            memcpy(_dataImage, _data, _numData*sizeof(T));
         };

   /// Constructor : no input data
   template <typename T, int V>
      Image<T,V>::Image(const int w, const int h) : 
         _width(w), _height(h),_numData(w*h*V), _sizeRow(w*V) {
            _dataImage = new T[_numData];
         }; 

   /// Constructor : input from a file
   template <typename T, int V>
      Image<T,V>::Image(const QString& fileName) {
         cout << fileName << " is being opened" << endl;
         QImage* image = new QImage(fileName);
         loadQImage(*image);
         delete(image);
      };

   /// Constructor : input from a QImage
   template <typename T, int V>
      Image<T,V>::Image(const QImage& image) {
         loadQImage(image);
      };

   /// Copy Constructor
   template <typename T, int V>
      Image<T,V>::Image(const Image<T,V>& image) : 
         _width(image._width),
         _height(image._height),
         _numData(image._numData),
         _sizeRow(image._sizeRow) {
            _dataImage = new T[_numData];
            memcpy(_dataImage,image._dataImage,_numData*sizeof(T));
         };

   /// Operator= : make a copy
   template<typename T, int V>
      inline void Image<T,V>::operator=(const Image<T,V>& image) {
         if (_numData != image._numData) {
            delete(_dataImage);
            _numData = image._numData;
            _dataImage = new T[_numData];
         };
         memcpy(_dataImage,image._dataImage,_numData*sizeof(T));
         _width = image._width;
         _height = image._height;
         _numData = image._numData;
         _sizeRow = image._sizeRow;
      };

   /// Operator/=
   template <typename T, int V>
      inline void Image<T,V>::operator/= (const Image<T,V>& image) {
         for (int i = 0; i<_numData; ++i) {
            if (fabs(image[i]) > EPSILON)
               _dataImage[i] /= image[i];
         }
      };
   
   /// Operator*=
   template <typename T, int V>
      inline void Image<T,V>::operator*= (const Image<T,V>& image) {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] *= image[i];
         }
      };

   /// Operator+=
   template <typename T, int V>
      inline void Image<T,V>::operator+= (const Image<T,V>& image) {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] += image[i];
         }
      };

   /// Operator-=
   template <typename T, int V>
      inline void Image<T,V>::operator-= (const Image<T,V>& image) {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] -= image[i];
         }
      };

   /// Operator=
   template <typename T, int V>
      inline void Image<T,V>::operator= (const T lambda) {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] = lambda;
         }
      };

   /// Operator/=
   template <typename T, int V>
      inline void Image<T,V>::operator/= (const T lambda) {
         const T invLambda = 1.0/lambda;
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] *= invLambda;
         }
      };

   /// Operator*=
   template <typename T, int V>
      inline void Image<T,V>::operator*= (const T lambda) {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] *= lambda;
         }
      };

   /// Operator+=
   template <typename T, int V>
      inline void Image<T,V>::operator+= (const T lambda) {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] += lambda;
         }
      };

   /// Operator-=
   template <typename T, int V>
      inline void Image<T,V>::operator-= (const T lambda) {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] -= lambda;
         }
      };

   /// Operator[]
   template <typename T, int V>
      inline T& Image<T,V>::operator[] (const int i) {
         assert(i>=0 && i<_numData);
         return _dataImage[i];
      };

   /// Operator[]
   template <typename T, int V>
      inline T Image<T,V>::operator[] (const int i) const {
         assert(i>=0 && i<_numData);
         return _dataImage[i];
      };
   /// Operator()
   template <typename T, int V>
      inline T& Image<T,V>::operator() (const int i, const int j) const {
         assert(i>=0 && i<_height && j>=0 && j<_width && V==1);
         return _dataImage[(i*_width+j)];
      };
   /// Operator()
   template <typename T, int V>
      inline T& Image<T,V>::operator() (const int i, const int j,
            const int k) const {
         assert(i>=0 && i<_height && j>=0 && j<_width && k>=0 && k<V);
         return _dataImage[V*(i*_width+j)+k];
      };

   /// return the max element of the image
   template <typename T, int V>
      inline T Image<T,V>::max() const {
         return _dataImage[idamax_(_numData,_dataImage,1)-1];
      };
 
   /// return the min element of the image
   template <typename T, int V>
      inline T Image<T,V>::min() const {
         return _dataImage[idamin_(_numData,_dataImage,1)-1];
      };
  
   /// replace each value by 1/value
   template <typename T, int V>
      inline void Image<T,V>::inv() {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] = 1.0/MAX(EPSILON_IMAGE,_dataImage[i]);
         }
      };

   /// uncenter the data
   template <typename T, int V>
      inline void Image<T,V>::unCenter() {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] += MEAN_VALUE;
         }
      };

   /// erase the image
   template <typename T, int V>
      inline void Image<T,V>::clear() {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] = T();
         }
      };

   /// clip the values from MIN_VALUE to MAX_VALUE
   template <typename T, int V>
       void Image<T,V>::clipData() {
          for (int i = 0; i< _numData; ++i) {
             _dataImage[i] = clip(_dataImage[i]);
          }
       };

   /// save the image in a file
   template <>
      bool ImageRGB::saveFile(const QString& fileName) const {
         if (!_dataImage) return false;
         QImage* image = new QImage(_width,_height,32);
         for (int j = 0; j<_height; ++j) {
            for (int i = 0; i<_width; ++i) {
               image->setPixel(i,j,qRgb(
                           static_cast<int>(clip(_dataImage[3*(j*_width+i)])+MEAN_VALUE),
                           static_cast<int>(clip(_dataImage[3*(j*_width+i)+1])+MEAN_VALUE),
                           static_cast<int>(clip(_dataImage[3*(j*_width+i)+2])+MEAN_VALUE)));
            }
         }
         bool res = image->save(fileName,"PNG");
         delete(image);
         image = NULL;
         return res;
      };

   /// save the image in a file
   template <>
      bool ImageGray::saveFile(const QString& fileName) const {
         if (!_dataImage) return false;
         QImage* image = new QImage(_width,_height,8,256);
         for (int i = 0; i<= 255; ++i) {
            image->setColor(i,qRgb(i,i,i));
         }
         for (int j = 0; j<_height; ++j) {
            for (int i = 0; i<_width; ++i) {
               image->setPixel(i,j,static_cast<uint>(clip(
                           _dataImage[j*_width+i])+MEAN_VALUE));
            }
         }
         bool res = image->save(fileName,"PNG");
         delete(image);
         image = NULL;
         return res;
      };

   /// for debbuging
   template <>
      bool Image<double,9>::saveFile(const QString& fileName) const {
         cerr << "Error save image<double,9>" << endl;
      };

   /// print some information about the image
   template <typename T, int V>
      void Image<T,V>::printInformation() const {
         switch (V) {
            case 1 : 
               cout << "Gray Image, " << _width << " x " << _height << endl;
               break;
            case 3 : 
               cout << "RGB Image, " << _width << " x " << _height << endl;
               break;
            default :
               cout << "Multi-Channel Image, " << _width << " x " << 
                  _height << endl;
         }
      };

   /// add a white gaussian noise to the image
   template <typename T, int V>
      void Image<T,V>::addNoiseGaussian(const T sigma) {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] += sigma*normalDistrib<T>();
         }
      };

   /// add a non-uniform white gaussian noise to the image
   template <typename T, int V>
      void Image<T,V>::addNoiseGaussian(const Image<T,V>& mapSigma) {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] += mapSigma[i]*normalDistrib<T>();
         }
      };

   /// add a non-uniform white gaussian noise to the image
   template <typename T, int V>
      void Image<T,V>::addNoiseGaussian(const QString& fileMapSigma) {
         Image<T,V> map(fileMapSigma);
         addNoiseGaussian(map);
      };

   /// apply a map for inpainting
   template <typename T, int V>
      void Image<T,V>::applyMap(const Image<T,V>& map) {
         for (int i = 0; i<_numData; ++i) {
            if (fabs(map[i] < EPSILON)) {
               _dataImage[i] = -MEAN_VALUE;
            }
         }
      };

   /// put a Bayer pattern in the image
   template <typename T, int V>
      void Image<T,V>::bayerPattern() {
         assert(V == 3);
         int count = 0;
         for (int i = 0; i< _height; ++i) {
            for (int j = 0; j< _width; ++j) {
               bool eq1 = (i % 2 == 0);
               bool eq2 = (j % 2 == 0);
               if (eq1 && !eq2) {
                  _dataImage[count++] = 1.0;
                  _dataImage[count++] = 0.0;
                  _dataImage[count++] = 0.0;
               } else if (eq2 && !eq1) {
                  _dataImage[count++] = 0.0;
                  _dataImage[count++] = 0.0;
                  _dataImage[count++] = 1.0;
               } else {
                  _dataImage[count++] = 0.0;
                  _dataImage[count++] = 1.0;
                  _dataImage[count++] = 0.0;
               }
            }
         }

      };

   /// mozaic the image
   template <typename T, int V>
      void Image<T,V>::mozaic() {
         assert(V == 3);
         Image<T,V> im(_width,_height);
         im.bayerPattern();
         applyMap(im);
      };

   /// copy the patch number num in the vector patch
   /// TODO : use the function changeStructurePatch to simplify
   /// this function
   template <typename T, int V>
      void inline Image<T,V>::setPatch(VectR& patch, const int num,
            const int sizeEdgePatches, const int level,
            const int offset) const {
         assert(level > 0 && sizeEdgePatches > 0 && offset >= 0);
         if (level == 1) {
            int count = 0;
            int offset1 = V*num;
            for (int i = 0; i<sizeEdgePatches; ++i) {
               int offset2 = i*_sizeRow;
               for (int j = 0; j< V*sizeEdgePatches; ++j) {
                  assert(offset1+offset2+j < _numData);
                  patch(count+offset) = _dataImage[offset1+offset2+j];
                  ++count;
               }
            }
         } else {
            int newSizeEdge = sizeEdgePatches/2;
            int sizePatches = V*newSizeEdge*newSizeEdge;
            int newOffset = offset;
            setPatch(patch,quad<true,true>(num,newSizeEdge),
                  newSizeEdge, level-1,newOffset);
            newOffset+=sizePatches;
            setPatch(patch,quad<false,true>(num,newSizeEdge),
                  newSizeEdge, level-1,newOffset);
            newOffset+=sizePatches;
            setPatch(patch,quad<true,false>(num,newSizeEdge),
                  newSizeEdge, level-1,newOffset);
            newOffset+=sizePatches;
            setPatch(patch,quad<false,false>(num,newSizeEdge),
                  newSizeEdge, level-1,newOffset);
         }
      };

   /// increment the value of one area, which corresponds to the patch
   /// i, with a n x n patch
   template <typename T, int V>
      inline void Image<T,V>::incrPatch(const int i, const int n) {
         int offset = V*i;
         for (int j = 0; j<n; ++j) {
            int offset2 = j*_sizeRow;
            for (int k = 0; k<n*V; ++k) {
               _dataImage[offset+offset2+k] += 1.0;
            }
         }
      };

   /// add the value of the patch, which corresponds to the patch
   /// i, with a n x n patch
   /// TODO : use the function reverseStructurePatch from Dictionary
   template <typename T, int V>
      inline void Image<T,V>::addPatch(const VectR& patch,
            const int numPatch, const int sizeEdgePatch, 
            const int level, const int offset) {
         assert(level > 0);
         assert(V*numPatch+(sizeEdgePatch-1)*_sizeRow +
               sizeEdgePatch*V-1 < _numData);
         if (level == 1) {
            int count = 0;
            int offset1 = V*numPatch;
            for (int i = 0; i<sizeEdgePatch; ++i) {
               int offset2 = i*_sizeRow;
               for (int j = 0; j<sizeEdgePatch*V; ++j) {
                  _dataImage[offset1+offset2+j] += patch(count+offset);
                  ++count;
               }
            }
         } else {
            int newSizeEdge = sizeEdgePatch/2;
            int sizePatches = V*newSizeEdge*newSizeEdge;
            int newOffset = offset;
            addPatch(patch,quad<true,true>(numPatch,newSizeEdge),
                  newSizeEdge, level-1,newOffset);
            newOffset+=sizePatches;
            addPatch(patch,quad<false,true>(numPatch,newSizeEdge),
                  newSizeEdge, level-1,newOffset);
            newOffset+=sizePatches;
            addPatch(patch,quad<true,false>(numPatch,newSizeEdge),
                  newSizeEdge, level-1,newOffset);
            newOffset+=sizePatches;
            addPatch(patch,quad<false,false>(numPatch,newSizeEdge),
                  newSizeEdge, level-1,newOffset);
         }
      };

   /// compute the PSNR with the given image
   template<typename T, int V>
      T Image<T,V>::PSNR(const Image<T,V>& input) const {
         T eqm = 0;
         T tmp;
         for (int i = 0; i< input._numData; ++i) {
            tmp = _dataImage[i]-input[i];
            eqm += tmp*tmp;
         }
         eqm /= static_cast<T>(input._numData);
         const T range = MAX_VALUE-MIN_VALUE;
         return 10.0*log(range*range/eqm)/log(10.0);
      };


   /// return a new patch index for lower scales
   template <typename T, int V>
      template <bool Top, bool Left>
      inline int Image<T,V>::quad(const int numPatch,
            const int newSizeEdge) const {
         int offset = 0;
         if (!Top) offset += newSizeEdge;
         if (!Left) offset += newSizeEdge*_width;
         return offset+numPatch;
      };

   /// binarize
   template <typename T, int V>
      void Image<T,V>::binarize() {
         for (int i = 0; i<_numData; ++i) {
            _dataImage[i] = fabs(_dataImage[i]) < EPSILON ?
               0.0 : 1.0;
         }
      };

   /// reduce the size without linear filtering
   template <typename T, int V>
      void Image<T,V>::reduceScale(const int scale) {
         _sizeRow /= scale;
         _height /= scale;
         _width /= scale;
         _numData /= scale;
         T* _newDataImage = new T[_numData];
         for (int i = 0; i<_height; ++i) {
            for (int j = 0; j<_width; ++j) {
               for (int k = 0; k<V; ++k) {
                  _newDataImage[V*(i*_width+j)+k] =
                     _dataImage[scale*V*(i*scale*_width+j)+k];
               }
            }
         }
         delete(_dataImage);
         _dataImage = _newDataImage;
      };

   /// TODO convolution by a (small) kernel (no FFT)
   template <typename T, int V>
      void Image<T,V>::convolution(const Matrix& kernel) {
         // TODO check if FFT is necessary etc...
      };

   /// convolution by a (small) separable kernel, K = a^T b (no FFT)
   template <typename T, int V>
      void Image<T,V>::convolutionSeparable(const VectR& a, const VectR& b) {
         // TODO
      };

   /// blurring
   template <typename T, int V>
      void Image<T,V>::blur(const QString& fileKernel) {
         ifstream f(fileKernel);
         int mode;
         if (!f) {
            cerr << "Failed to open kernel file" << endl;
            exit(1);
         }
         f >> mode;
         f.close();
         if (mode == 0) {
            Matrix mat;
            openKernel<T>(fileKernel,mat);
            convolution(mat);
         } else if (mode == 1) {
            VectR A,B;
            openKernelSep<T>(fileKernel,A,B);
            convolutionSeparable(A,B);
         } else {
            cerr << "general transform not implemented" << endl;
            exit(1);
         }
      };

   /// load a color image from a QImage
   template <>
      void ImageRGB::loadQImage(const QImage& image) {
         _width = image.width();
         _height = image.height();
         _sizeRow = _width*3;
         _numData = 3*_width*_height;
         _dataImage = new double[_numData];
         if (image.depth() != 32) {
            cerr << "loadQImage, bad parameter" << endl;
            exit(1);
         }
         int count = 0;
         for (int j = 0; j<_height; ++j) {
            for (int i = 0; i<_width; ++i) {
               int pixel = image.pixel(i,j);
               _dataImage[count] = qRed(pixel) - MEAN_VALUE;
               ++count;
               _dataImage[count] = qGreen(pixel) - MEAN_VALUE;
               ++count;
               _dataImage[count] = qBlue(pixel) - MEAN_VALUE;
               ++count;
            }
         }
      };

   /// load a gray image from a QImage
   template <>
      void ImageGray::loadQImage(const QImage& image) {
         _width = image.width();
         _height = image.height();
         _numData = _width*_height;
         _sizeRow = _width;
         _dataImage = new double[_numData];
         uchar* pixel = image.bits();
         if (image.depth() == 8) {
            for (int i = 0; i<_width; ++i) {
               for (int j = 0; j<_height; ++j) {
                  _dataImage[j*_width+i] = image.pixelIndex(i,j)
                     - MEAN_VALUE;
               }
            }
         } else if (image.depth() == 32) {
            int count = 0;
            for (int j = 0; j<_height; ++j) {
               for (int i = 0; i<_width; ++i) {
                  _dataImage[count] = qGray(image.pixel(i,j)) - 
                     MEAN_VALUE;
                  ++count;
               }
            }
         } else {
            cerr << "loadQImage, bad parameter" << endl;
            exit(1);
         }
      };

/*   template<typename T, int V>
      Image<T,V>* Image<T,V>::quad(int i) {
         Image<T,V>* res = new Image<T,V>(_width/2,_height/2);
         int offset1,offset2;
         switch (i) {
            case 0: offset1 = offset2 = 0; break;
            case 1: offset1 = _width/2; offset2 = 0; break;
            case 2: offset1 = 0; offset2 = _height/2; break;
            default : offset1 = _width/2; offset2 = _height/2;
         }
         for (int i = 0; i<_width/2; ++i) {
            for (int j = 0; j<_height/2; ++j) {
               for (int k = 0; k<V; ++k) {
                  (*res)(i,j,k) = (*this)(i+offset1,j+offset2,k);
               }               
            }
         }
         return res;
      };
   
   template<typename T, int V>
      void Image<T,V>::setQuad(const Image<T,V>& im, int i) {
         int offset1,offset2;
         switch (i) {
            case 0: offset1 = offset2 = 0; break;
            case 1: offset1 = _width/2; offset2 = 0; break;
            case 2: offset1 = 0; offset2 = _height/2; break;
            default : offset1 = _width/2; offset2 = _height/2;
         }
         for (int i = 0; i<_width/2; ++i) {
            for (int j = 0; j<_height/2; ++j) {
               for (int k = 0; k<V; ++k) {
                  (*this)(i+offset1,j+offset2,k)= im(i,j,k);
               }               
            }
         }
      };*/

   /// extract region of interest
   template<typename T, int V>
      Image<T,V>* Image<T,V>::regionOfInterest(int i, int j, 
            int sizeX, int sizeY) const {
         Image<T,V>* res = new Image<T,V>(sizeY,sizeX);
         for (int ii = 0; ii<sizeX; ++ii) {
            for (int jj = 0; jj<sizeY; ++jj) {
               for (int kk = 0; kk<V; ++kk) {
                  (*res)(ii,jj,kk) = (*this)(ii+i,jj+j,kk);
               }
            }
         }
         return res;
      };

   /// add im to a region of interest
   template<typename T, int V>
      void Image<T,V>::addToRegionOfInterest(const Image<T,V>& im,
            int i, int j) {
         for (int ii = 0; ii<im.height(); ++ii) {
            for (int jj = 0; jj<im.width(); ++jj) {
               for (int kk = 0; kk<V; ++kk) {
                  (*this)(ii+i,jj+j,kk) += im(ii,jj,kk);
               }
            }
         }
      };

   /// extract channel
   template<typename T, int V>
      Image<T,1>* Image<T,V>::channel(int k) const {
         Image<T,1>* res = new Image<T,1>(_width,_height);
         for (int i = 0; i<_height; ++i) {
            for (int j = 0; j<_width; ++j) {
               (*res)(i,j) = (*this)(i,j,k);
            }
         }
         return res;
      };

   /// set channel
   template<typename T, int V>
      void Image<T,V>::setChannel(const Image<T,1>& im,
            int k) {
         for (int i = 0; i<_height; ++i) {
            for (int j = 0; j<_width; ++j) {
               (*this)(i,j,k) = im(i,j);
            }
         }
      };

};

#endif // IMAGE_H
