/* inpainting.h */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
*/

/*!
 * \file
 *                Multiscale KSVD Inpainting
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File inpainting.h
 * \brief Contains the class Inpainting */

#ifndef INPAINTING_H
#define INPAINTING_H

#include "denoising.h"

using namespace std;

namespace KSVD {

   /*! 
    * namespace for weighted K-SVD algorithms
    * */
   namespace Weighted {

      template <typename T, int V> class Inpainting : 
         public KSVD::Algo<T,V> {
            public:
               /// Constructor
               Inpainting(const Image<T,V>* const input, 
                     const Image<T,V>* const map,  
                     const int J, const QString& fileDictionary,
                     const int maxL = DEFAULT_MAX_SPARSITY_INPAINTING,
                     const Image<T,V>* const original = NULL,
                     const int overlapFactor = OVERLAP_FACTOR);
               Inpainting(const Image<T,V>* const input, 
                     const Image<T,V>* const map,  
                     const int J, Dictionary<T,V>* dictionary,
                     const int maxL = DEFAULT_MAX_SPARSITY_INPAINTING,
                     const Image<T,V>* const original = NULL,
                     const int overlapFactor = OVERLAP_FACTOR);
               /// Destructor
               virtual ~Inpainting() { delete(_map); }

               /// print some information about the algorithm's parameters
               virtual void printInformation() const;
               /// execute the denoising algorithm
               virtual void exec(const T gam = 0.0, const bool weightConstant = true);
               /// execute the denoising algorithm
               virtual void fastExec(const T gam = 0.0);
               /// execute the denoising algorithm
               virtual void blockExec(const T gam = 0.0, const bool fast = true);

            protected:
               /// generate the output
               void generateOutput(list<int>& listPatches);

               /// matrix \f$ \beta \f$:w
               Image<T,V>* _map;
               /// number of iterations
               int _J;
               /// maximum sparsity
               int _maxL;
         };

      /// Constructor
      template <typename T, int V>
         Inpainting<T,V>::Inpainting(const Image<T,V>* const input, 
               const Image<T,V>* const map, const int J, 
               const QString& fileDictionary, 
               const int maxL,
               const Image<T,V>* const original,
               const int overlapFactor) :
            Algo<T,V>(input,fileDictionary,original,overlapFactor),
            _maxL(maxL),
            _J(J) {
               if (map == NULL) {
                  _map = NULL;
               } else {
                  _map = new Image<T,V>(*map);
                  _map->binarize();
               }
               this->_N = this->_dictionary->N(); 
            };

      /// Alternative constructor
      template <typename T, int V>
         Inpainting<T,V>::Inpainting(const Image<T,V>* const input, 
               const Image<T,V>* const map, const int J, 
               Dictionary<T,V>* dictionary, 
               const int maxL,
               const Image<T,V>* const original,
               const int overlapFactor) :
            Algo<T,V>(input,dictionary,original,overlapFactor),
            _maxL(maxL),
            _J(J) {
               if (map == NULL) {
                  _map = NULL;
               } else {
                  _map = new Image<T,V>(*map);
                  _map->binarize();
               }
               this->_N = this->_dictionary->N(); 
            };

      /// print some information about the algorithm's parameters.
      template <typename T, int V>
         void Inpainting<T,V>::printInformation() const {
            cout << "Executing inpainting Algorithm with "
               << "Multiscale KSVD" << endl;
            cout << "J = " << this->_J << endl;
            cout << "L = " << _maxL << endl;
            Algo<T,V>::printInformation();
         };


      /// execute the inpainting algorithm.
      template <typename T, int V> 
         void Inpainting<T,V>::exec(const T gam, const bool weightConstant) {
            list<int>* listPatches = new list<int>();
#ifdef _OPENMP
            /// adapt the number of threads to the number of processors
            /// available
            int NUM_THREADS =  omp_get_num_procs();
#else 
            int NUM_THREADS = 1;
#endif
            int numPatches = this->generateListPatches(*listPatches);
            /// if RGB space is selected, use the RGB metric for denoising
            /// this boolean is related to a future functionality 
            /// which will reduce drastically the memory occupation
            bool storeResidual = this->_J != 0 && (numPatches*
                  this->_dictionary->n(this->_dictionary->N()-1)) <
               MAX_ALLOCATE_RESIDUAL;
            /// storeResidual functionality has not been implemented yet
            /// so it is set to true by default
            storeResidual = true;
            /// allocate the workspace memory for the algorithm
            this->_dictionary->allocateLearning(numPatches,this->_N,true,
                  storeResidual);
            cout << "Iteration n0" << endl;
            /// one class Processor for each thread
            KSVD::Processor<T,V>** processors = 
               new KSVD::Processor<T,V>*[NUM_THREADS];
            for (int i = 0; i<NUM_THREADS; ++i) {
               processors[i] = new KSVD::Weighted::Processor<T,V>(
                     this->_dictionary, EPSILON_IMAGE, this->_N,
                     93,0.0,this->_input,_map,_maxL, V <= 3);
            }
            Timer clock;
            clock.start();
            this->approximation(processors,*listPatches,NUM_THREADS,
                  V <= 1);
            this->_dictionary->printSparsity();
            /// Principal loop
            for (int i = 1; i<=this->_J; ++i) {
               clock.pause();
               /// Do not take into account the time spent for 
               /// statistical analysis
               generateOutput(*listPatches);
               this->printPSNR();
               clock.restart();

               /// go back to the algorithm
               cout << "Iteration n" << i << endl;
               for (int j = 0; j<ITER_WEIGHTED_LEARNING; ++j)
                  this->_dictionary->performLearning(false,false,
                        this->_input,_map);
               this->approximation(processors,*listPatches,NUM_THREADS,
                     V <= 1);
               this->_dictionary->printSparsity();
            }

            generateOutput(*listPatches);
            this->printPSNR();
            /// deallocate the workspace for the learning
            this->_dictionary->terminateLearning();
            for (int i = 0; i<NUM_THREADS; ++i) {
               delete(processors[i]);
               processors[i] = NULL;
            }
            delete[](processors);

            clock.stop();
            clock.printTimeElapsed();
            delete(listPatches);
            listPatches = NULL;
         };

      /// execute the inpainting algorithm.
      template <typename T, int V> 
         void Inpainting<T,V>::blockExec(const T gam, const bool fast) {
            Timer clock;
            clock.start();
            int tmpX = this->_input->height()-SIZE_BLOCK_DENOISING;
            int tmpY = this->_input->width()-SIZE_BLOCK_DENOISING;
            int sizeBlock = SIZE_BLOCK_DENOISING-
               this->_dictionary->sizeEdgeBigPatches();

            int numBlockX,numBlockY,sizeBlockDenoisingX,sizeBlockDenoisingY;
            if (tmpX < 0) {
               numBlockX = 1;
               sizeBlockDenoisingX = this->_input->height();
            } else {
               numBlockX = tmpX/sizeBlock+1;
               if (!(tmpX % sizeBlock == 0)) ++numBlockX;
               sizeBlockDenoisingX = SIZE_BLOCK_DENOISING;
            }
            if (tmpY < 0) {
               numBlockY = 1;
               sizeBlockDenoisingY = this->_input->width();
            } else {
               numBlockY = tmpY/sizeBlock+1;
               if (!(tmpY % sizeBlock == 0)) ++numBlockY;
               sizeBlockDenoisingY = SIZE_BLOCK_DENOISING;
            }

            delete(this->_output);
            this->_output = new Image<T,V>(this->_input->width(),
                  this->_input->height());
            *this->_output = 0.0;
            Image<T,V>* weights = new Image<T,V>(*this->_output);
            Image<T,V>* blockOnes = new Image<T,V>(sizeBlockDenoisingY,
                  sizeBlockDenoisingX);
            *blockOnes = 1.0;
            int count = 0;

            for (int i = 0; i<numBlockX; ++i) {
               for (int j = 0; j<numBlockY; ++j) {
                  cout << "Block " << count << " is being processed" << endl;
                  int cornerX = MAX(0,MIN(i*sizeBlock,
                           this->_input->height()-sizeBlockDenoisingX));
                  int cornerY = MAX(0,MIN(j*sizeBlock,
                           this->_input->width()-sizeBlockDenoisingY));
                  Image<T,V>* blockInput = this->_input->regionOfInterest(
                        cornerX,cornerY,
                        sizeBlockDenoisingX,sizeBlockDenoisingY);
                  Image<T,V>* blockMap = this->_map->regionOfInterest(
                        cornerX,cornerY,
                        sizeBlockDenoisingX,sizeBlockDenoisingY);
                  Inpainting<T,V>* algo = new Inpainting<T,V>(blockInput,blockMap,
                        count == 0 ? _J : _J/FACTOR_ITER_PROPAG_DICT
                        ,this->_dictionary,_maxL,NULL,this->_overlapFactor);
                  if (fast) {
                     algo->fastExec(gam);
                  } else {
                     algo->exec(gam);
                  }
                  this->_output->addToRegionOfInterest(algo->output(),
                        cornerX,cornerY);
                  weights->addToRegionOfInterest(*blockOnes,cornerX,cornerY);
                  algo->setDictionaryNull();
                  delete(algo);
                  delete(blockInput);
                  delete(blockMap);
                  ++count;
               }
            }
            *(this->_output) /= *weights;
            delete(weights);
            delete(blockOnes);
            clock.stop();
            clock.printTimeElapsed();
         };

      /// execute the inpainting algorithm.
      template <typename T, int V> 
         void Inpainting<T,V>::fastExec(const T gam) {
            list<int>* listPatches = new list<int>();
#ifdef _OPENMP
            /// adapt the number of threads to the number of processors
            /// available
            int NUM_THREADS =  omp_get_num_procs();
#else 
            int NUM_THREADS = 1;
#endif
            int numPatches = this->generateListPatches(*listPatches);
            /// if RGB space is selected, use the RGB metric for denoising
            /// this boolean is related to a future functionality 
            /// which will reduce drastically the memory occupation
            bool storeResidual = this->_J != 0 && (numPatches*
                  this->_dictionary->n(this->_dictionary->N()-1)) <
               MAX_ALLOCATE_RESIDUAL;
            /// storeResidual functionality has not been implemented yet
            /// so it is set to true by default
            storeResidual = true;
            /// one class Processor for each thread
            KSVD::Processor<T,V>** processors = 
               new KSVD::Processor<T,V>*[NUM_THREADS];
            for (int i = 0; i<NUM_THREADS; ++i) {
               processors[i] = new KSVD::Weighted::Processor<T,V>(
                     this->_dictionary, EPSILON_IMAGE, this->_N,
                     93,0.0,this->_input,_map,_maxL, V <= 3);
            }
            /// Learning step
            cout << "Learning on a random subset of patches" << endl;
            Timer clock;
            clock.start();

            /// Principal loop
            for (int i = 1; i<=this->_J; ++i) {
               /// allocate the workspace memory for the algorithm
               cout << "Iteration n" << i << endl;
               list<int>* prunedListPatches = new list<int>(*listPatches);
               pruneListPatches(*prunedListPatches, PRUNE_RATIO_DENOISING);
               int prunedNumPatches = prunedListPatches->size();

               this->_dictionary->allocateLearning(prunedNumPatches,this->_N,true,
                     storeResidual);
               this->approximation(processors,*prunedListPatches,NUM_THREADS, V <= 3);

               for (int j = 0; j<ITER_WEIGHTED_LEARNING; ++j) {
                  this->_dictionary->performLearning(false,false,
                        this->_input,_map);
               }
               this->_dictionary->printSparsity();
               this->_dictionary->terminateLearning();
               delete(prunedListPatches);
            }

            cout << "Reconstruction" << endl;
            this->_dictionary->allocateLearning(numPatches,this->_N,
                  true, storeResidual);
            this->approximation(processors,*listPatches,NUM_THREADS, V <= 3);
            this->_dictionary->printSparsity();

            generateOutput(*listPatches);
            this->printPSNR();
            /// deallocate the workspace for the learning
            this->_dictionary->terminateLearning();
            for (int i = 0; i<NUM_THREADS; ++i) {
               delete(processors[i]);
               processors[i] = NULL;
            }
            delete[](processors);

            clock.stop();
            clock.printTimeElapsed();
            delete(listPatches);
            listPatches = NULL;
         };


      /// generate the output inpainted image
      template <typename T, int V> 
         void Inpainting<T,V>::generateOutput(list<int>& listPatches) {
            KSVD::Algo<T,V>::generateOutput(listPatches,0.0,_map);
            for (int i = 0; i<_map->numData(); ++i) {
               if ((*_map)[i] == 1.0) 
                  (*this->_output)[i] = (*this->_input)[i];
            }
         };

      /// inpainting with double learning
      template <typename T, int V> class InpaintingDoubleLearning : 
         public Inpainting<T,V> {
            public:
               /// Constructor
               InpaintingDoubleLearning(const Image<T,V>* const input, 
                     const Image<T,V>* const map,  
                     const int J, const QString& fileDictionary,
                     const int maxL = DEFAULT_MAX_SPARSITY_INPAINTING,
                     const Image<T,V>* const original = NULL,
                     const int Llearning = DEFAULT_LOW_SPARSITY,
                     const bool doubleLearning = true,
                     const int overlapFactor = OVERLAP_FACTOR);

               /// print some information about the algorithm's parameters
               virtual void printInformation() const;
               /// execute the denoising algorithm
               virtual void exec(const T gam = 0.0);

            private:
               QString _fileDictionary;
               /// sparsity factor for double learning
               int _Llearning;
               /// double learning or not
               bool _doubleLearning;
         };

      /// Constructor
      template <typename T, int V>
         InpaintingDoubleLearning<T,V>::InpaintingDoubleLearning(
               const Image<T,V>* const input, 
               const Image<T,V>* const map,  
               const int J, const QString& fileDictionary,
               const int maxL,
               const Image<T,V>* const original,
               const int Llearning,
               const bool doubleLearning,
               const int overlapFactor) :
            Inpainting<T,V>(input,map,J,fileDictionary,maxL,original),
            _Llearning(Llearning),
            _doubleLearning(doubleLearning),
            _fileDictionary(fileDictionary) { };

      /// print some information about the algorithm's parameters.
      template <typename T, int V>
         void InpaintingDoubleLearning<T,V>::printInformation() const {
            cout << "Executing inpainting Algorithm with "
               << "Multiscale KSVD and Double Learning" << endl;
            cout << "J = " << this->_J << endl;
            cout << "L = " << this->_maxL << endl;
            cout << "Llearning = " <<_Llearning << endl;
            Algo<T,V>::printInformation();
         };

      /// execute the inpainting algorithm.
      template <typename T, int V> 
         void InpaintingDoubleLearning<T,V>::exec(const T gam) {
            int J = this->_J;
            if (_doubleLearning) J = this->_J/2;
            this->_J = 0;
            // First reconstruction
            Inpainting<T,V>::exec(gam);
            cout << "First reconstruction achieved" << endl;
            if (J > 0) {
               KSVD::Denoising<T,V>* algo;
               Dictionary<T,V>* originalDictionary = 
                  new Dictionary<T,V>(*(this->_dictionary));
               // First learning with low sparsity
               if (_doubleLearning) {
                  algo = new KSVD::Denoising<T,V>(this->_output,
                        EPSILON_IMAGE,J,_fileDictionary,NULL,_Llearning);
                  algo->setFixedSparsity(true);
                  algo->setStats(false);
                  algo->setN(this->_dictionary->N());
                  algo->setChoiceC(93);
                  algo->exec(gam,false);
                  cout << "First Learning achieved" << endl;
                  // Second reconstruction
                  delete(this->_dictionary);
                  this->_dictionary = new Dictionary<T,V>(*originalDictionary,
                        algo->dictionary());
                  delete(algo);
                  Inpainting<T,V>::exec(gam);
                  cout << "Second Reconstruction achieved" << endl;
               }
               algo = new KSVD::Denoising<T,V>(this->_output,
                     EPSILON_IMAGE,J,_fileDictionary,NULL,_Llearning);
               algo->setFixedSparsity(true);
               algo->setStats(false);
               algo->setN(this->_dictionary->N());
               algo->setChoiceC(93);
               algo->exec(gam,false);
               cout << "Second Learning achieved" << endl;
               delete(this->_dictionary);
               this->_dictionary = new Dictionary<T,V>(*originalDictionary,
                     algo->dictionary());
               delete(algo);
               delete(originalDictionary);
               Inpainting<T,V>::exec(gam);
               cout << "Third Reconstruction achieved" << endl;
            }
            if (_doubleLearning) this->_J = 2*J;
         };
   };

   /// Denoising algorithm for grayscale images
   typedef KSVD::Weighted::Inpainting<double,1> InpaintingGray;
   /// Denoising algorithm for RGB images
   typedef KSVD::Weighted::Inpainting<double,3> InpaintingRGB;

   /// Denoising algorithm for grayscale images
   typedef KSVD::Weighted::InpaintingDoubleLearning<double,1> 
      InpaintingDoubleLearningGray;
   /// Denoising algorithm for RGB images
   typedef KSVD::Weighted::InpaintingDoubleLearning<double,3> 
      InpaintingDoubleLearningRGB;

   namespace Sequence {

      template <typename T = double, int V = 1, int W = 3> class 
         Inpainting : public Algo<T,V,W> {
            public:
               /// Constructor
               Inpainting(const Video<T,V,W>* const input, 
                     Video<T,V,W>* map,  
                     const int J, const QString& fileDictionary,
                     const int maxL = 
                     DEFAULT_MAX_SPARSITY_INPAINTING_VIDEO,
                     const Video<T,V,W>* const original = NULL,
                     const int overlapFactor = OVERLAP_FACTOR_VIDEO );
               /// Destructor
               virtual ~Inpainting() { }

               /// print some information about the algorithm's parameters
               virtual void printInformation() const;
               /// execute the denoising algorithm
               void exec(const T gam = 0.0);

            protected:
               /// matrix \f$ \beta \f$:w
               Video<T,V,W>* _map;
               /// number of iterations
               int _J;
               /// maximum sparsity
               int _maxL;
         }; 

      /// Constructor
      template <typename T, int V, int W>
         Inpainting<T,V,W>::Inpainting(const Video<T,V,W>* const input, 
               Video<T,V,W>* map,  
               const int J, const QString& fileDictionary,
               const int maxL,
               const Video<T,V,W>* const original,
               const int overlapFactor) :
            Algo<T,V,W>(input,fileDictionary,original,overlapFactor),
            _maxL(maxL),
            _map(map),
            _J(J) {
               this->_N = this->_dictionary->N(); 
            };

      /// print some information about the algorithm's parameters.
      template <typename T, int V, int W>
         void Inpainting<T,V,W>::printInformation() const {
            cout << "Executing inpainting Algorithm with "
               << "Multiscale KSVD for Video" << endl;
            cout << "J = " << this->_J << endl;
            cout << "L = " << _maxL << endl;
            Algo<T,V,W>::printInformation();
         };

      /// execute the inpainting algorithm.
      template <typename T, int V, int W> 
         void Inpainting<T,V,W>::exec(const T gam) {
            this->_output = new Video<T,V,W>(this->_input->width(),
                  this->_input->height(), this->_input->nFrames());
            this->_output->clear();
            const int sizeFrame = this->_input->width()*
               this->_input->height()*V;
            const int offset = W/2;
            KSVD::Weighted::Inpainting<T,V*W>* algorithm;
            for (int i = offset;
                  i< this->_input->nFrames()-offset; ++i) {
               cout << "********************************************"
                  << endl;
               cout << "Frame #" << i << " is being processed" << endl;
               const_cast<Video<T,V,W>*>(
                     this->_input)->setCurrentFrame(i);
               const_cast<Video<T,V,W>*>(
                     _map)->setCurrentFrame(i);
               Image<T,V*W>* original = NULL;
               if (this->_original) {
                  const_cast<Video<T,V,W>*>(
                        this->_original)->setCurrentFrame(i);
                  original = this->_original->currentFrames();
               }
               algorithm = new KSVD::Weighted::Inpainting<T,V*W>(
                     this->_input->currentFrames(),_map->currentFrames(), 
                     i == offset ? NUM_ITER_VIDEO_INITIAL : _J, 
                     this->_dictionary,_maxL,
                     original,
                     this->_overlapFactor);
               algorithm->exec(gam);

               const Image<T,V*W>& output = algorithm->output();
               for (int j = 0; j<W; ++j) {
                  for (int l = 0; l<this->_input->width()*
                        this->_input->height(); ++l) {
                     for (int k = 0; k<V; ++k) {
                        (*this->_output)[(i-offset+j)*sizeFrame+l*V+k] +=
                           output[W*V*l+j*V+k];
                     }
                  }
               }
               algorithm->setDictionaryNull();
               delete(algorithm);
               algorithm = NULL;
            }
            cout << "********************************************"
               << endl;
            /// finish the averaging
            int boundary = MIN(W-1,this->_input->nFrames()-W);
            for (int i = 1; i<boundary; ++i) {
               for (int j = 0; j<sizeFrame; ++j) {
                  (*this->_output)[i*sizeFrame+j] /= (i+1);
               }
            }
            for (int i = boundary; i<this->_input->nFrames()-boundary; ++i) {
               for (int j = 0; j<sizeFrame; ++j) {
                  (*this->_output)[i*sizeFrame+j] /= boundary+1;
               }
            }
            for (int i = this->_input->nFrames()-boundary; 
                  i<this->_input->nFrames()-1; ++i) {
               for (int j = 0; j<sizeFrame; ++j) {
                  (*this->_output)[i*sizeFrame+j] /= 
                     (this->_input->nFrames()-i);
               }
            }
            /// adjust final values using the map
            this->_output->restoreFromMap(*_map,*this->_input);
         };

      /// inpainting with double learning
      template <typename T, int V, int W> class InpaintingDoubleLearning : 
         public Inpainting<T,V,W> {
            public:
               /// Constructor
               InpaintingDoubleLearning(const Video<T,V,W>* const input, 
                     Video<T,V,W>* map,  
                     const int J, const QString& fileDictionary,
                     const int maxL = 
                     DEFAULT_MAX_SPARSITY_INPAINTING_VIDEO,
                     const Video<T,V,W>* const original = NULL,
                     const int Llearning = (DEFAULT_LOW_SPARSITY+V),
                     const int overlapFactor = OVERLAP_FACTOR_VIDEO );

               /// print some information about the algorithm's parameters
               virtual void printInformation() const;
               /// execute the inpainting algorithm
               void exec(const T gam = 0.0, const bool fast = false,
                     const bool block = false);

            private:
               QString _fileDictionary;
               /// sparsity factor for double learning
               int _Llearning;
         };

      /// Constructor
      template <typename T, int V, int W>
         InpaintingDoubleLearning<T,V,W>::InpaintingDoubleLearning(
               const Video<T,V,W>* const input, 
               Video<T,V,W>* map,  
               const int J, const QString& fileDictionary,
               const int maxL,
               const Video<T,V,W>* const original,
               const int Llearning,
               const int overlapFactor) : 
            Inpainting<T,V,W>(input,map,J,fileDictionary,maxL,original,
                  overlapFactor),
            _Llearning(Llearning),
            _fileDictionary(fileDictionary) {  };

      /// print some information about the algorithm's parameters
      template <typename T, int V, int W>
         void InpaintingDoubleLearning<T,V,W>::printInformation() const {
            cout << "Executing inpainting Algorithm with "
               << "Multiscale KSVD for Video and double learning" << endl;
            cout << "J = " << this->_J << endl;
            cout << "L1 = " << this->_maxL << endl;
            cout << "L2 = " << _Llearning << endl;
            Algo<T,V,W>::printInformation();
         };

      /// execute the inpainting algorithm
      template <typename T, int V, int W>
         void InpaintingDoubleLearning<T,V,W>::exec(const T gam, const bool fast,
               const bool block) {
            this->_output = new Video<T,V,W>(this->_input->width(),
                  this->_input->height(), this->_input->nFrames());
            this->_output->clear();
            const int sizeFrame = this->_input->width()*
               this->_input->height()*V;
            const int offset = W/2;
            KSVD::Weighted::InpaintingDoubleLearning<T,V*W>* algorithm;
            for (int i = offset;
                  i< this->_input->nFrames()-offset; ++i) {
               cout << "********************************************"
                  << endl;
               cout << "Frame #" << i << " is being processed" << endl;
               const_cast<Video<T,V,W>*>(
                     this->_input)->setCurrentFrame(i);
               const_cast<Video<T,V,W>*>(
                     this->_map)->setCurrentFrame(i);
               Image<T,V*W>* original = NULL;
               if (this->_original) {
                  const_cast<Video<T,V,W>*>(
                        this->_original)->setCurrentFrame(i);
                  original = this->_original->currentFrames();
               }
               algorithm = new KSVD::Weighted::InpaintingDoubleLearning<T,V*W>(
                     this->_input->currentFrames(),this->_map->currentFrames(), 
                     this->_J, 
                     this->_fileDictionary,this->_maxL,
                     original, DEFAULT_LOW_SPARSITY+V, false, 
                     this->_overlapFactor);
               if (block) {
                  algorithm->blockExec(gam,fast);
               } else {
                  if (fast) {
                     algorithm->fastExec(gam);
                  } else {
                     algorithm->exec(gam);
                  }
               }


               const Image<T,V*W>& output = algorithm->output();
               for (int j = 0; j<W; ++j) {
                  for (int l = 0; l<this->_input->width()*
                        this->_input->height(); ++l) {
                     for (int k = 0; k<V; ++k) {
                        (*this->_output)[(i-offset+j)*sizeFrame+l*V+k] +=
                           output[W*V*l+j*V+k];
                     }
                  }
               }
               algorithm->setDictionaryNull();
               delete(algorithm);
               algorithm = NULL;
            }
            cout << "********************************************"
               << endl;
            /// finish the averaging
            int boundary = MIN(W-1,this->_input->nFrames()-W);
            for (int i = 1; i<boundary; ++i) {
               for (int j = 0; j<sizeFrame; ++j) {
                  (*this->_output)[i*sizeFrame+j] /= (i+1);
               }
            }
            for (int i = boundary; i<this->_input->nFrames()-boundary; ++i) {
               for (int j = 0; j<sizeFrame; ++j) {
                  (*this->_output)[i*sizeFrame+j] /= boundary+1;
               }
            }
            for (int i = this->_input->nFrames()-boundary; 
                  i<this->_input->nFrames()-1; ++i) {
               for (int j = 0; j<sizeFrame; ++j) {
                  (*this->_output)[i*sizeFrame+j] /= 
                     (this->_input->nFrames()-i);
               }
            }
            /// adjust final values using the map
            this->_output->restoreFromMap(*this->_map,*this->_input);

         };
   };

   typedef KSVD::Sequence::Inpainting<double,1,3> InpaintingVideoGray;
   typedef KSVD::Sequence::Inpainting<double,1,5> InpaintingVideoHQGray;
   typedef KSVD::Sequence::Inpainting<double,3,3> InpaintingVideoRGB;
   typedef KSVD::Sequence::Inpainting<double,3,5> InpaintingVideoHQRGB;
   typedef KSVD::Sequence::InpaintingDoubleLearning<double,1,3> 
      InpaintingDoubleLearningVideoGray;
   typedef KSVD::Sequence::InpaintingDoubleLearning<double,1,5> 
      InpaintingDoubleLearningHQVideoGray;
   typedef KSVD::Sequence::InpaintingDoubleLearning<double,3,3> 
      InpaintingDoubleLearningVideoRGB;
   typedef KSVD::Sequence::InpaintingDoubleLearning<double,3,5> 
      InpaintingDoubleLearningHQVideoRGB;

};

#endif // INPAINTING_H
