/* main.cpp */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright 2006-2007 Julien Mairal 

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

/*!
 * \file
 *           Main file for Multiscale KSVD Denoising
 *                         
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                  File main.cpp
 * \brief Main file
 *
 * \mainpage KSVDRestore, MultiScale KSVD
 *
 * \author Julien Mairal
 *
 * \date January, 2007
 *
 *                   Multiscale KSVD for :
 *                   - Denoising
 *                   - InPainting
 *                   - Demosaicing
 *                   - Video Denoising
 *                   - Video InPainting
 *                   - Video Demosaicing
 *
 * \section intro_sec Introduction
 *    - This software is parallel implementation of the three following papers,
 *         -# M. Elad and M. Aharon, "Image Denoising Via Sparse and 
 *           Redundant representations over Learned Dictionaries", the 
 *           IEEE Trans. on Image Processing, Vol. 15, no. 12, pp. 3736-3745,
 *           December 2006. 
 *         -# J. Mairal, M. Elad, and G. Sapiro, "Sparse Representation for 
 *           Color Image Restoration", Submitted to IEEE Trans. on Image 
 *           Processing.
 *           http://www.ima.umn.edu/preprints/oct2006/2139.pdf
 *         -# J. Mairal, G. Sapiro and M. Elad, "Multiscale sparse image 
 *           representation with learned dictionaries". submitted.
 *           http://www.ima.umn.edu/preprints/jan2007/2152.pdf
 *    - The first paper describes the first K-SVD denoising algorithm. 
 *    - The second paper describes the inpainting/denoising/demosaicing for
 *    color image restoration.
 *    - The last paper presents a multiscale version of the first denoising
 *    algorithm as well as some algorithmic improvements.
 *
 * \section license_sec License and usage rules
 *    - This software is distributed under the GPL license.
 *      
 * \section install_sec Installation
 * \subsection libraries_sec Required libraries
 *     - This software requires linking to QT 3.*, 
 *     http://www.trolltech.com/products/qt
 *     - This software requires an BLAS/LAPACK library. It is compatible with
 *     any libraries which provide such an interface.
 *     It as been tested with:
 *           -# BLAS/LAPACK 
 *           http://www.netlib.org/lapack/, http://netlib.caspur.it/blas/
 *           -# Intel Math Kernel Library 
 *           recommended for maximum performance
 *     - This software uses a modified version of parts of the following free
 *     libraries:
 *           -# TNT (http://math.nist.gov/tnt/)
 *           -# CPPLAPACK (http://cpplapack.sourceforge.net/)
 *           Installing these libraries is not necessary since they are 
 *           embedded in the software.
 *         
 * \subsection arch_sec Supported platforms
*     It supports currently Linux/Unix platforms.
*
* \subsection compile_sec Supported compilers
*     GCC 4.* (http://gcc.gnu.org/) and ICC (Intel compiler)
*     At the moment the multi-threading capabilities require OpenMP to be 
*     supported by the compiler.
*
* \subsection Compilation
*     - Edit KSVDRestore.pro, follow the instructions inside 
*     - qmake-qt3
*     - make
*
* \section use_sec Utilization
*     The command ./KSVDRestore gives a brief summary
*     To use some routines in your code, just look in the file main.cpp
*/

#include "video.h"
#include "denoising.h"
#include "inpainting.h"
#include "demosaicing.h"
#include "trainer.h"
#include <qdir.h>


/// namespace for all the KSVD algorithm
using namespace KSVD;

/// brief summary of the usage
static void printUsage() {
   cout << "KSVDRestoration, multi-scale implementation of the " << endl;
   cout << "K-SVD algorithm implementation,  Copyright (C) 2007 "
      << "Julien Mairal, julien.mairal@m4x.org";
   cout << endl << endl;
   cout << "Usage : " << endl;
   cout << "./KSVD_Restore -denoising cleanOriginalImage dictionary ..." << endl;
   cout << "... outputImage sigma J" << endl << endl;
   cout << "./KSVD_Restore -videoDenoising cleanOriginalVideoFolder Nimages dictionary ..." << endl;
   cout << "... outputVideoFolder sigma J" << endl << endl;
   cout << "./KSVD_Restore -hqVideoDenoising cleanOriginalVideoFolder Nimages dictionary ..." << endl;
   cout << "... outputVideoFolder sigma J" << endl << endl;
   cout << "./KSVD_Restore -denoisingNonUniform cleanOriginalImage dictionary ..." << endl;
   cout << "... outputImage imageSigma J" << endl << endl;
   cout << "./KSVD_Restore -inpainting cleanOriginalImage dictionary ..." << endl;
   cout << "... outputImage mapHoles J L" << endl << endl;
   cout << "./KSVD_Restore -inpaintingDouble cleanOriginalImage dictionary ..." << endl;
   cout << "... outputImage mapHoles J L1 L2" << endl << endl;
   cout << "./KSVD_Restore -inpaintingVideo cleanOriginalVideoFolder dictionary ..." << endl;
   cout << "... outputVideoFolder Nimages percentToRemove J L" << endl << endl;
   cout << "./KSVD_Restore -inpaintingVideoHQ cleanOriginalVideoFolder dictionary ..." << endl;
   cout << "... outputVideoFolder Nimages percentToRemove J L" << endl << endl;
   cout << "./KSVD_Restore -demosaicing cleanOriginalImage dictionary ..." << endl;
   cout << "... outputImage J L" << endl << endl;
   cout << "./KSVD_Restore -demosaicingVideo inputFolder dictionary ..." << endl;
   cout << "... outputFolder Nimages J L" << endl << endl;
   cout << "./KSVD_Restore -demosaicingVideoHQ inputFolder dictionary ..." << endl;
   cout << "... outputFolder Nimages J L" << endl << endl;
   cout << "./KSVD_Restore -deblurring cleanOriginalImage dictionary ..." << endl;
   cout << "... outputImage pathToKernel J L1 L2" << endl << endl;
   cout << "./KSVD_Restore -training -RGB/GRAY pathToDataBase initialDictionary ...";
   cout << endl;
   cout << "... outputDictionary J L numImages numPatchesPerImage";
   cout << "./KSVD_Restore -trainingVideo -RGB/GRAY fileDataBase initialDictionary ...";
   cout << endl;
   cout << "... outputDictionary J L numPatchesPerImage";
   cout << endl << endl;
   cout << "Other usages have not been implemented yet" << endl;
};

/// main function
int main(int argc, char** argv) {
#ifdef NRANDOM
   srand(0);
   seed = -1;
#else
   srand( (unsigned) time(NULL));
   seed = - abs(time(NULL));
#endif

   WEIGHT_CONSTANT=2.5;
   try {
      if (argc < 2) { printUsage(); exit(0); }

      QString function(argv[1]);

      if (QString::compare(function,"-denoising") == 0 ||
            QString::compare(function,"-fdenoising") == 0 ||
            QString::compare(function,"-blockFastDenoising") == 0 ||
            QString::compare(function,"-blockDenoising") == 0) {
         // denoising algorithm is chosen
         if (argc != 13) {
            throw ("BAD PARAMETERS");
         }
         bool block = QString::compare(function,"-blockFastDenoising") == 0 ||
            QString::compare(function,"-blockDenoising") == 0;
         bool fast = QString::compare(function,"-fdenoising") == 0 || 
            QString::compare(function,"-blockFastDenoising") == 0;
         const double sigma = QString(argv[5]).toDouble();
         WEIGHT_CONSTANT = QString(argv[9]).toDouble();
         PRUNE_RATIO_DENOISING = QString(argv[10]).toDouble();
         const double gamma = QString(argv[11]).toDouble();
         const QString fileOutput(argv[4]);
         const QString fileDictionary(argv[3]);
         const int J = QString(argv[6]).toInt();
         SIZE_BLOCK_DENOISING = QString(argv[12]).toInt();
         NN = QString(argv[7]).toInt();
         choiceC = QString(argv[8]).toInt();
         const QImage inputQImage = QImage(argv[2]);
         cout << "file : " << argv[2] << " is being opened" << endl;

         // test is Gray space or RGB space is chosen
         if (inputQImage.isGrayscale()) {
            // GRAY denoising algorithm
            ImageGray* originalImage = new ImageGray(inputQImage);
            ImageGray* inputImage = new ImageGray(*originalImage);
            // create the noisy image
            DenoisingGray* algorithm;
            if (sigma > 0) {
               inputImage->addNoiseGaussian(sigma);
               algorithm = new DenoisingGray(inputImage,
                     sigma, J, fileDictionary, originalImage);
            } else {
               ifstream f("noise.tmp");
               for (int i = 0; i<inputImage->width()*inputImage->height();
                     ++i) {
                  double noise;
                  f >> noise;
                  (*inputImage)[i] += noise;
               }
               f.close();
               algorithm = new DenoisingGray(inputImage,
                     -sigma, J, fileDictionary);
            }
            algorithm->printInformation();
            if (block) {
               algorithm->blockExec(gamma,fast);
            } else {
               if (fast) {
                  algorithm->fastExec(gamma);
               } else {
                  algorithm->exec(gamma);
               }
            } 
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            algorithm = NULL;
            delete(inputImage);
            inputImage = NULL;
            delete(originalImage);
            originalImage = NULL;
         } else {
            // RGB denoising algorithm
            ImageRGB* originalImage = new ImageRGB(inputQImage);
            ImageRGB* inputImage = new ImageRGB(*originalImage);
            // create the noisy image
            DenoisingRGB* algorithm;
            if (sigma > 0) {
               inputImage->addNoiseGaussian(sigma);
               algorithm = new DenoisingRGB(inputImage,
                     sigma, J, fileDictionary, originalImage);
            } else {
               algorithm = new DenoisingRGB(inputImage,
                     -sigma, J, fileDictionary);
            }
            algorithm->printInformation();
            if (block) {
               algorithm->blockExec(gamma,fast);
            } else {
               if (fast) {
                  algorithm->fastExec(gamma);
               } else {
                  algorithm->exec(gamma);
               }
            } 
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            algorithm = NULL;
            delete(inputImage);
            inputImage = NULL;
            delete(originalImage);
            originalImage = NULL;
         }
      } else if (QString::compare(function,"-RGBperChanneldenoising") == 0) {
         // denoising algorithm is chosen
         if (argc != 13) {
            throw ("BAD PARAMETERS");
         }
         bool block = true;
         bool fast = true;
         const double sigma = QString(argv[5]).toDouble();
         WEIGHT_CONSTANT = QString(argv[9]).toDouble();
         PRUNE_RATIO_DENOISING = QString(argv[10]).toDouble();
         const double gamma = QString(argv[11]).toDouble();
         const QString fileOutput(argv[4]);
         const QString fileDictionary(argv[3]);
         const int J = QString(argv[6]).toInt();
         SIZE_BLOCK_DENOISING = QString(argv[12]).toInt();
         NN = QString(argv[7]).toInt();
         choiceC = QString(argv[8]).toInt();
         const QImage inputQImage = QImage(argv[2]);
         cout << "file : " << argv[2] << " is being opened" << endl;

         // test is Gray space or RGB space is chosen
         // RGB denoising algorithm
         ImageRGB* originalImage = new ImageRGB(inputQImage);
         ImageRGB* inputImage = new ImageRGB(*originalImage);
         ImageRGB* outputImage = new ImageRGB(*originalImage);
         outputImage->clear();
         // create the noisy image
         inputImage->addNoiseGaussian(sigma);
         inputImage->saveFile("test.png");
         for (int k = 0; k<3; ++k) {
            cout << "Processing channel " << k << endl;
            ImageGray* channel = inputImage->channel(k);
            DenoisingGray* algorithm = new DenoisingGray(channel,
                  sigma, J, fileDictionary, NULL);
            algorithm->printInformation();
            algorithm->blockExec(gamma,fast);
            outputImage->setChannel(algorithm->output(),k);
            delete(algorithm);
            delete(channel);
         }
         cout << "PSNR : " << outputImage->PSNR(*originalImage) << endl;
         delete(outputImage);
         delete(inputImage);
         delete(originalImage);
      } else if (QString::compare(function,"-videoDenoising") == 0) {
         // denoising algorithm is chosen
         if (argc != 12) {
            throw ("BAD PARAMETERS");
         }
         const double sigma = QString(argv[6]).toDouble();
         const QString outputFolder(argv[5]);
         const QString fileDictionary(argv[4]);
         const int J = QString(argv[7]).toInt();
         const QString inputFolder(argv[2]);
         const int Nimages = QString(argv[3]).toInt();
         WEIGHT_CONSTANT = QString(argv[8]).toDouble();
         const double gam = QString(argv[9]).toDouble();
         NN = QString(argv[10]).toInt();
         choiceC = QString(argv[11]).toInt();
         cout << "input folder : " << argv[2] << " is being opened" << endl;
         /// a video is input folder 
         QString firstImage(inputFolder+"0.png");
         const QImage inputQImage(firstImage);

         // test is Gray space or RGB space is chosen
         if (inputQImage.isGrayscale()) {
            // GRAY denoising algorithm
            VideoGray* originalVideo = new VideoGray(inputFolder,Nimages);
            //            originalVideo->save(outputFolder+"/init");
            VideoGray* inputVideo = new VideoGray(*originalVideo);
            // create the noisy image
            inputVideo->addNoiseGaussian(sigma);
            DenoisingVideoGray* algorithm = new DenoisingVideoGray(inputVideo,
                  sigma, J, fileDictionary, originalVideo);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
            if (!algorithm->save(outputFolder)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            algorithm = NULL;
            delete(inputVideo);
            inputVideo = NULL;
            delete(originalVideo);
            originalVideo = NULL;
         } else {
            // RGB denoising algorithm
            VideoRGB* originalVideo = new VideoRGB(inputFolder,Nimages);
            VideoRGB* inputVideo = new VideoRGB(*originalVideo);
            // create the noisy image
            inputVideo->addNoiseGaussian(sigma);
            DenoisingVideoRGB* algorithm = new DenoisingVideoRGB(inputVideo,
                  sigma, J, fileDictionary, originalVideo);
            algorithm->printInformation();
            algorithm->exec(gam);
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
            if (!algorithm->save(outputFolder)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            algorithm = NULL;
            delete(inputVideo);
            inputVideo = NULL;
            delete(originalVideo);
            originalVideo = NULL;
         }
      } else if (QString::compare(function,"-hqVideoDenoising") == 0) {
         // denoising algorithm is chosen
         if (argc != 12) {
            throw ("BAD PARAMETERS");
         }
         const double sigma = QString(argv[6]).toDouble();
         const QString outputFolder(argv[5]);
         const QString fileDictionary(argv[4]);
         const int J = QString(argv[7]).toInt();
         const QString inputFolder(argv[2]);
         const int Nimages = QString(argv[3]).toInt();
         WEIGHT_CONSTANT = QString(argv[8]).toDouble();
         const double gam = QString(argv[9]).toDouble();
         NN = QString(argv[10]).toInt();
         choiceC = QString(argv[11]).toInt();

         cout << "input folder : " << argv[2] << " is being opened" << endl;
         /// a video is input folder 
         QString firstImage(inputFolder+"0.png");
         const QImage inputQImage(firstImage);

         // test is Gray space or RGB space is chosen
         if (inputQImage.isGrayscale()) {
            // GRAY denoising algorithm
            VideoHQGray* originalVideo = new VideoHQGray(inputFolder,Nimages);
            //            originalVideo->save(outputFolder+"/init");
            VideoHQGray* inputVideo = new VideoHQGray(*originalVideo);
            // create the noisy image
            inputVideo->addNoiseGaussian(sigma);
            DenoisingHQVideoGray* algorithm = new DenoisingHQVideoGray(inputVideo,
                  sigma, J, fileDictionary, originalVideo);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
            if (!algorithm->save(outputFolder)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            algorithm = NULL;
            delete(inputVideo);
            inputVideo = NULL;
            delete(originalVideo);
            originalVideo = NULL;
         } else {
            // RGB denoising algorithm
            VideoHQRGB* originalVideo = new VideoHQRGB(inputFolder,Nimages);
            VideoHQRGB* inputVideo = new VideoHQRGB(*originalVideo);
            // create the noisy image
            inputVideo->addNoiseGaussian(sigma);
            DenoisingHQVideoRGB* algorithm = new DenoisingHQVideoRGB(inputVideo,
                  sigma, J, fileDictionary, originalVideo);
            algorithm->printInformation();
            algorithm->exec(gam);
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
            if (!algorithm->save(outputFolder)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            algorithm = NULL;
            delete(inputVideo);
            inputVideo = NULL;
            delete(originalVideo);
            originalVideo = NULL;
         }
      } else if (QString::compare(function,"-denoisingNonUniform") == 0) {
         // denoising non uniform mode
         if (argc != 7) {
            throw ("BAD PARAMETERS");
         }
         const QString sigmaFile(argv[5]);
         const QString fileOutput(argv[4]);
         const QString fileDictionary(argv[3]);
         const int J = QString(argv[6]).toInt();
         const QImage inputQImage = QImage(argv[2]);
         cout << "file : " << argv[2] << " is being opened" << endl;

         // test is Gray space or RGB space is chosen
         if (inputQImage.isGrayscale()) {
            // non uniform GRAY denoising
            ImageGray* originalImage = new ImageGray(inputQImage);
            ImageGray* inputImage = new ImageGray(*originalImage);
            ImageGray* sigmaMap = new ImageGray(sigmaFile);
            sigmaMap->unCenter();
            // create the noisy image
            inputImage->addNoiseGaussian(*sigmaMap);
            Weighted::DenoisingGray* algorithm = 
               new Weighted::DenoisingGray(inputImage,
                     *sigmaMap, J, fileDictionary, originalImage);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(sigmaMap);
            delete(algorithm);
            delete(inputImage);
            delete(originalImage);
         } else {
            // non uniform RGB denoising
            ImageRGB* originalImage = new ImageRGB(inputQImage);
            ImageRGB* inputImage = new ImageRGB(*originalImage);
            ImageRGB* sigmaMap = new ImageRGB(sigmaFile);
            sigmaMap->unCenter();
            // create the noisy image
            inputImage->addNoiseGaussian(*sigmaMap);
            inputImage->saveFile("test.png");
            Weighted::DenoisingRGB* algorithm = 
               new Weighted::DenoisingRGB(inputImage, *sigmaMap, 
                     J, fileDictionary, originalImage);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(sigmaMap);
            delete(algorithm);
            delete(inputImage);
            delete(originalImage);
         }
      } else if (QString::compare(function,"-denoisingNonUniformN") == 0) {
         // denoising non uniform mode
         if (argc != 8) {
            throw ("BAD PARAMETERS");
         }
         const QString sigmaFile(argv[5]);
         const double gam = QString(argv[7]).toDouble();
         const QString fileOutput(argv[4]);
         const QString fileDictionary(argv[3]);
         const int J = QString(argv[6]).toInt();
         const QImage inputQImage = QImage(argv[2]);
         cout << "file : " << argv[2] << " is being opened" << endl;

         // test is Gray space or RGB space is chosen
         if (inputQImage.isGrayscale()) {
            // non uniform GRAY denoising
            ImageGray* originalImage = new ImageGray(inputQImage);
            ImageGray* sigmaMap = new ImageGray(sigmaFile);
            sigmaMap->unCenter();
            // create the noisy image
            Weighted::DenoisingGray* algorithm = 
               new Weighted::DenoisingGray(originalImage,
                     *sigmaMap, J, fileDictionary);
            algorithm->printInformation();
            algorithm->exec(gam);
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(sigmaMap);
            delete(algorithm);
            delete(originalImage);
         } else {
            // non uniform RGB denoising
            ImageRGB* originalImage = new ImageRGB(inputQImage);
            ImageRGB* sigmaMap = new ImageRGB(sigmaFile);
            sigmaMap->unCenter();
            // create the noisy image
            Weighted::DenoisingRGB* algorithm = 
               new Weighted::DenoisingRGB(originalImage, *sigmaMap, 
                     J, fileDictionary);
            algorithm->printInformation();
            algorithm->exec(gam);
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(sigmaMap);
            delete(algorithm);
            delete(originalImage);
         }
      } else if (QString::compare(function,"-inpainting") == 0 ||
            QString::compare(function,"-blockFastInpainting") == 0) {
         // inpainting mode
         if (argc != 11) {
            throw ("BAD PARAMETERS");
         }
         bool block = QString::compare(function,"-blockFastInpainting") == 0;
         const QString fileOutput(argv[4]);
         const QString fileDictionary(argv[3]);
         const int J = QString(argv[6]).toInt();
         const int L = QString(argv[7]).toInt();
         const QImage inputQImage(argv[2]);
         const QString mapFile(argv[5]);
         WEIGHT_CONSTANT = QString(argv[8]).toDouble();
         PRUNE_RATIO_DENOISING = QString(argv[9]).toDouble();
         SIZE_BLOCK_DENOISING = QString(argv[10]).toInt();

         // test is Gray space or RGB space is chosen
         if (inputQImage.isGrayscale()) {
            // GRAY inpainting
            ImageGray* originalImage = new ImageGray(inputQImage);
            ImageGray* inputImage = new ImageGray(*originalImage);
            ImageGray* mapImage = new ImageGray(mapFile);
            mapImage->unCenter();
            /// create the inpainted image
            inputImage->applyMap(*mapImage);
            InpaintingGray* algorithm = new InpaintingGray(inputImage,
                  mapImage, J, fileDictionary,L,originalImage);
            algorithm->printInformation();
            if (block) {
               algorithm->blockExec(0.0,true);
            } else {
               algorithm->exec();
            }
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(mapImage);
            delete(algorithm);
            delete(inputImage);
            delete(originalImage);
         } else {
            // RGB inpainting
            ImageRGB* originalImage = new ImageRGB(inputQImage);
            ImageRGB* inputImage = new ImageRGB(*originalImage);
            ImageRGB* mapImage = new ImageRGB(mapFile);
            mapImage->unCenter();
            // create the inpainted image
            inputImage->applyMap(*mapImage);
            InpaintingRGB* algorithm = new InpaintingRGB(inputImage,
                  mapImage, J, fileDictionary,L,originalImage);
            algorithm->printInformation();
            if (block) {
               algorithm->blockExec(0.0,true);
            } else {
               algorithm->exec();
            }
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(mapImage);
            delete(algorithm);
            delete(inputImage);
            delete(originalImage);
         }
      } else if (QString::compare(function,"-inpaintingDouble") == 0) {
         // inpainting mode with learning on estimated image
         if (argc != 9) {
            throw ("BAD PARAMETERS");
         }
         const QString fileOutput(argv[4]);
         const QString fileDictionary(argv[3]);
         const int J = QString(argv[6]).toInt();
         const int L = QString(argv[7]).toInt();
         const int L2 = QString(argv[8]).toInt();
         const QImage inputQImage(argv[2]);
         const QString mapFile(argv[5]);

         // test is Gray space or RGB space is chosen
         if (inputQImage.isGrayscale()) {
            // GRAY inpainting with learning on estimated image 
            ImageGray* originalImage = new ImageGray(inputQImage);
            ImageGray* inputImage = new ImageGray(*originalImage);
            ImageGray* mapImage = new ImageGray(mapFile);
            mapImage->unCenter();
            // create the inpainted image
            inputImage->applyMap(*mapImage);
            InpaintingDoubleLearningGray* algorithm = 
               new InpaintingDoubleLearningGray(inputImage,
                     mapImage, J, fileDictionary,L,originalImage,L2);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(mapImage);
            delete(algorithm);
            delete(inputImage);
            delete(originalImage);
         } else {
            // RGB inpainting with learning on estimated image 
            ImageRGB* originalImage = new ImageRGB(inputQImage);
            ImageRGB* inputImage = new ImageRGB(*originalImage);
            ImageRGB* mapImage = new ImageRGB(mapFile);
            mapImage->unCenter();
            // create the inpainted image
            inputImage->applyMap(*mapImage);
            InpaintingDoubleLearningRGB* algorithm = new 
               InpaintingDoubleLearningRGB(inputImage,
                     mapImage, J, fileDictionary,L,originalImage,L2);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(fileOutput);
            if (!algorithm->saveFile(fileOutput)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(mapImage);
            delete(algorithm);
            delete(inputImage);
            delete(originalImage);
         }
      } else if (QString::compare(function,"-inpaintingVideo") == 0) {
         // inpainting mode
         if (argc != 10) {
            throw ("BAD PARAMETERS");
         }
         const QString inputFolder(argv[2]);
         const QString fileDictionary(argv[3]);
         const QString outputFolder(argv[4]);
         const int Nimages = QString(argv[5]).toInt();
         const int percent = QString(argv[6]).toInt();
         const int J = QString(argv[7]).toInt();
         const int L = QString(argv[8]).toInt();
         WEIGHT_CONSTANT = QString(argv[9]).toDouble();
         cout << "input folder : " << inputFolder 
            << " is being opened" << endl;
         /// a video is input folder 
         QString firstImage(inputFolder+"0.png");
         const QImage inputQImage(firstImage);

         // test is Gray space or RGB space is chosen
         if (inputQImage.isGrayscale()) {
            // GRAY denoising algorithm
            VideoGray* originalVideo = 
               new VideoGray(inputFolder,Nimages);
            VideoGray* inputVideo = new VideoGray(*originalVideo);
            VideoGray* map = inputVideo->removePercents(percent);
            InpaintingVideoGray* algorithm = new InpaintingVideoGray(
                  inputVideo,map,J,fileDictionary,L,originalVideo);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
            if (!algorithm->save(outputFolder)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            delete(inputVideo);
            delete(originalVideo);
            delete(map);
         } else {
            // RGB denoising algorithm
            VideoRGB* originalVideo = new VideoRGB(inputFolder,Nimages);
            VideoRGB* inputVideo = new VideoRGB(*originalVideo);
            VideoRGB* map = inputVideo->removePercents(percent);
            InpaintingVideoRGB* algorithm = new InpaintingVideoRGB(
                  inputVideo,map,J,fileDictionary,L,originalVideo);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
            if (!algorithm->save(outputFolder)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            delete(inputVideo);
            delete(originalVideo);
            delete(map);
         }
      } else if (QString::compare(function,"-inpaintingVideoHQ") == 0) {
         // inpainting mode
         if (argc != 10) {
            throw ("BAD PARAMETERS");
         }
         const QString inputFolder(argv[2]);
         const QString fileDictionary(argv[3]);
         const QString outputFolder(argv[4]);
         const int Nimages = QString(argv[5]).toInt();
         const int percent = QString(argv[6]).toInt();
         const int J = QString(argv[7]).toInt();
         const int L = QString(argv[8]).toInt();
         WEIGHT_CONSTANT = QString(argv[9]).toDouble();
         cout << "input folder : " << inputFolder 
            << " is being opened" << endl;
         /// a video is input folder 
         QString firstImage(inputFolder+"0.png");
         const QImage inputQImage(firstImage);

         // test is Gray space or RGB space is chosen
         if (inputQImage.isGrayscale()) {
            // GRAY denoising algorithm
            VideoHQGray* originalVideo = 
               new VideoHQGray(inputFolder,Nimages);
            VideoHQGray* inputVideo = new VideoHQGray(*originalVideo);
            VideoHQGray* map = inputVideo->removePercents(percent);
            InpaintingVideoHQGray* algorithm = new InpaintingVideoHQGray(
                  inputVideo,map,J,fileDictionary,L,originalVideo);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
            if (!algorithm->save(outputFolder)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            delete(inputVideo);
            delete(originalVideo);
            delete(map);
         } else {
            // RGB denoising algorithm
            VideoHQRGB* originalVideo = new VideoHQRGB(inputFolder,Nimages);
            VideoHQRGB* inputVideo = new VideoHQRGB(*originalVideo);
            VideoHQRGB* map = inputVideo->removePercents(percent);
            InpaintingVideoHQRGB* algorithm = new InpaintingVideoHQRGB(
                  inputVideo,map,J,fileDictionary,L,originalVideo);
            algorithm->printInformation();
            algorithm->exec();
            algorithm->printPSNR();
            algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
            if (!algorithm->save(outputFolder)) {
               throw("Unknown Error, no output from the algorithm");
            }
            delete(algorithm);
            delete(inputVideo);
            delete(originalVideo);
            delete(map);
         }
      } else if (QString::compare(function,"-demosaicing") == 0) {
         // Demosaicing mode
         if (argc != 7) {
            throw ("BAD PARAMETERS");
         }
         // TODO : fix WEIGHT_CONSTANT
         const QString fileOutput(argv[4]);
         const QString fileDictionary(argv[3]);
         const int J = QString(argv[5]).toInt();
         const int L = QString(argv[6]).toInt();
         const QImage inputQImage(argv[2]);
         ImageRGB* originalImage = new ImageRGB(inputQImage);
         ImageRGB* inputImage = new ImageRGB(*originalImage);
         inputImage->mozaic();
         Demosaicing* algorithm = new 
            Demosaicing(inputImage, J, fileDictionary,L,originalImage);
         algorithm->printInformation();
         algorithm->exec();
         algorithm->printPSNR();
         algorithm->writeDictionaryToImage(fileOutput);
         if (!algorithm->saveFile(fileOutput)) {
            throw("Unknown Error, no output from the algorithm");
         }
         delete(algorithm);
         delete(inputImage);
         delete(originalImage);
      } else if (QString::compare(function,"-demosaicingVideo") == 0) {
         // Demosaicing mode
         if (argc != 8) {
            throw ("BAD PARAMETERS");
         }
         const QString inputFolder(argv[2]);
         const QString fileDictionary(argv[3]);
         const QString outputFolder(argv[4]);
         const int Nimages = QString(argv[5]).toInt();
         const int J = QString(argv[6]).toInt();
         const int L = QString(argv[7]).toInt();

         VideoRGB* originalVideo = new VideoRGB(inputFolder,Nimages);
         VideoRGB* inputVideo = new VideoRGB(*originalVideo);
         inputVideo->mozaic();
         DemosaicingVideo* algorithm = new 
            DemosaicingVideo(inputVideo, J, fileDictionary,L,originalVideo);
         algorithm->printInformation();
         algorithm->exec();
         algorithm->printPSNR();
         algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
         if (!algorithm->save(outputFolder)) {
            throw("Unknown Error, no output from the algorithm");
         }
         delete(algorithm);
         delete(inputVideo);
         delete(originalVideo);
      } else if (QString::compare(function,"-demosaicingVideoHQ") == 0) {
         // Demosaicing mode
         if (argc != 8) {
            throw ("BAD PARAMETERS");
         }
         const QString inputFolder(argv[2]);
         const QString fileDictionary(argv[3]);
         const QString outputFolder(argv[4]);
         const int Nimages = QString(argv[5]).toInt();
         const int J = QString(argv[6]).toInt();
         const int L = QString(argv[7]).toInt();

         VideoHQRGB* originalVideo = new VideoHQRGB(inputFolder,Nimages);
         VideoHQRGB* inputVideo = new VideoHQRGB(*originalVideo);
         inputVideo->mozaic();
         DemosaicingVideoHQ* algorithm = new 
            DemosaicingVideoHQ(inputVideo, J, fileDictionary,L,originalVideo);
         algorithm->printInformation();
         algorithm->exec();
         algorithm->printPSNR();
         algorithm->writeDictionaryToImage(outputFolder+"/dico.png");
         if (!algorithm->save(outputFolder)) {
            throw("Unknown Error, no output from the algorithm");
         }
         delete(algorithm);
         delete(inputVideo);
         delete(originalVideo);
      } else if (QString::compare(function,"-training") == 0) {
         // Training a dictionary on an image database
         if (argc != 10) {
            throw ("BAD PARAMETERS");
         }
         const QString colorSpace(argv[2]);
         const QString pathToDataBase(argv[3]);
         const QString initialDictionary(argv[4]);
         const QString outputDictionary(argv[5]);
         const int J = QString(argv[6]).toInt();
         const int L = QString(argv[7]).toInt();
         const int numImages = QString(argv[8]).toInt();
         const int numPatchesPerImage = QString(argv[9]).toInt();

         if (QString::compare(colorSpace,"-GRAY")==0) {
            // gray color space
            TrainerGray* trainer = new TrainerGray(initialDictionary,
                  outputDictionary,numPatchesPerImage,L,J);
            for (int i = 1; i<=numImages; ++i) {
               QString image(pathToDataBase);
               image += QString::number(i);
               image += ".png";
               trainer->addToDatabase(image);
            }
            trainer->run();
            delete(trainer);
         } else if (QString::compare(colorSpace,"-RGB")==0) {
            // RGB color space
            TrainerRGB* trainer = new TrainerRGB(initialDictionary,
                  outputDictionary,numPatchesPerImage,L,J);
            for (int i = 1; i<=numImages; ++i) {
               QString image(pathToDataBase);
               image += QString::number(i);
               image += ".png";
               trainer->addToDatabase(image);
            }
            trainer->run();
            delete(trainer);
         } else {
            throw ("BAD PARAMETERS");
         }
         cout << "End of the training" << endl;
      } else if (QString::compare(function,"-trainingVideo") == 0) {
         // Training a dictionary on a video database
         if (argc != 9) {
            throw ("BAD PARAMETERS");
         }
         const QString colorSpace(argv[2]);
         const QString baseName(argv[3]);
         const QString initialDictionary(argv[4]);
         const QString outputDictionary(argv[5]);
         const int J = QString(argv[6]).toInt();
         const int L = QString(argv[7]).toInt();
         const int numPatchesPerImage = QString(argv[8]).toInt();

         if (QString::compare(colorSpace,"-GRAY")==0) {
            // gray color space
            TrainerVideoGray* trainer = new TrainerVideoGray(
                  initialDictionary,
                  outputDictionary,numPatchesPerImage,L,J);
            QDir d(baseName);
            d.setFilter(QDir::Dirs || QDir::NoSymLinks);
            const QFileInfoList *list = d.entryInfoList();
            QFileInfoListIterator it( *list );
            QFileInfo *fi;
            while ( (fi = it.current()) != 0 ) {
               if ((QString::compare(fi->fileName(),".") != 0) &&
                     (QString::compare(fi->fileName(),"..") != 0)) {
                  QDir dd(d);
                  dd.cd(fi->fileName());
                  dd.setFilter(QDir::Files);
                  int numImagesFolder = dd.count();
                  trainer->addToDatabase(fi->filePath()+"/image",
                        numImagesFolder);
               }
               ++it;
            }
            trainer->run();
            delete(trainer);
         } else if (QString::compare(colorSpace,"-RGB")==0) {
            // RGB color space
            TrainerVideoRGB* trainer = new TrainerVideoRGB(
                  initialDictionary,
                  outputDictionary,numPatchesPerImage,L,J);
            QDir d(baseName);
            d.setFilter(QDir::Dirs || QDir::NoSymLinks);
            const QFileInfoList *list = d.entryInfoList();
            QFileInfoListIterator it( *list );
            QFileInfo *fi;
            while ( (fi = it.current()) != 0 ) {
               if ((QString::compare(fi->fileName(),".") != 0) &&
                     (QString::compare(fi->fileName(),"..") != 0)) {
                  QDir dd(d);
                  dd.cd(fi->fileName());
                  dd.setFilter(QDir::Files);
                  int numImagesFolder = dd.count();
                  trainer->addToDatabase(fi->filePath()+"/image",
                        numImagesFolder);
               }
               ++it;
            }

            trainer->run();
            delete(trainer);
         } else {
            throw("BAD PARAMETERS");
         }
      } else if (QString::compare(function,"-trainingVideoHQ") == 0) {
         // Training a dictionary on a video database
         if (argc != 9) {
            throw ("BAD PARAMETERS");
         }
         const QString colorSpace(argv[2]);
         const QString baseName(argv[3]);
         const QString initialDictionary(argv[4]);
         const QString outputDictionary(argv[5]);
         const int J = QString(argv[6]).toInt();
         const int L = QString(argv[7]).toInt();
         const int numPatchesPerImage = QString(argv[8]).toInt();

         if (QString::compare(colorSpace,"-GRAY")==0) {
            // gray color space
            TrainerVideoHQGray* trainer = new TrainerVideoHQGray(
                  initialDictionary,
                  outputDictionary,numPatchesPerImage,L,J);
            QDir d(baseName);
            d.setFilter(QDir::Dirs || QDir::NoSymLinks);
            const QFileInfoList *list = d.entryInfoList();
            QFileInfoListIterator it( *list );
            QFileInfo *fi;
            while ( (fi = it.current()) != 0 ) {
               if ((QString::compare(fi->fileName(),".") != 0) &&
                     (QString::compare(fi->fileName(),"..") != 0)) {
                  QDir dd(d);
                  dd.cd(fi->fileName());
                  dd.setFilter(QDir::Files);
                  int numImagesFolder = dd.count();
                  trainer->addToDatabase(fi->filePath()+"/image",
                        numImagesFolder);
               }
               ++it;
            }
            trainer->run();
            delete(trainer);
         } else if (QString::compare(colorSpace,"-RGB")==0) {
            // RGB color space
            TrainerVideoHQRGB* trainer = new TrainerVideoHQRGB(
                  initialDictionary,
                  outputDictionary,numPatchesPerImage,L,J);
            QDir d(baseName);
            d.setFilter(QDir::Dirs || QDir::NoSymLinks);
            const QFileInfoList *list = d.entryInfoList();
            QFileInfoListIterator it( *list );
            QFileInfo *fi;
            while ( (fi = it.current()) != 0 ) {
               if ((QString::compare(fi->fileName(),".") != 0) &&
                     (QString::compare(fi->fileName(),"..") != 0)) {
                  QDir dd(d);
                  dd.cd(fi->fileName());
                  dd.setFilter(QDir::Files);
                  int numImagesFolder = dd.count();
                  trainer->addToDatabase(fi->filePath()+"/image",
                        numImagesFolder);
               }
               ++it;
            }

            trainer->run();
            delete(trainer);
         } else {
            throw("BAD PARAMETERS");
         }
      } else if (QString::compare(function,"-external") == 0) {
         if (argc != 9) {
            throw ("BAD PARAMETERS");
         }
         const QString fileNameDictionary(argv[2]);
         const QString fileNameOutputDictionary(argv[3]);
         const QString fileNameData(argv[4]);
         const int numData = QString(argv[5]).toInt();
         const int J = QString(argv[6]).toInt();
         const int L = QString(argv[7]).toInt();
         const int V = QString(argv[8]).toInt();
         if (V==1) {
            TrainerGray* trainer = new TrainerGray(fileNameDictionary,
                  fileNameOutputDictionary,
                  fileNameData,numData,L,J);
            trainer->run();
            delete(trainer);
         } else if (V==3) {
            TrainerRGB* trainer = new TrainerRGB(fileNameDictionary,
                  fileNameOutputDictionary,
                  fileNameData,numData,L,J);
            trainer->run();
            delete(trainer);
         } else {
            throw ("BAD PARAMETERS");
         }
      } else if (QString::compare(function,"-help") == 0) {
         // brief summary of the usage
         printUsage();         
      } else {
         throw("BAD PARAMETERS");
      }
      exit(0);
   } catch (const char* s) {
      if (QString::compare(s,"BAD PARAMETERS") == 0) {
         cerr << "Check your parameters" << endl;
         printUsage();
         exit(1);
      }
      if (QString::compare(s,"NOT IMPLEMENTED")== 0) {
         cerr << "You are calling a function which has not been " 
            << "implemented yet." << endl;
         printUsage();
         exit(1);
      }
      cerr << "Exception during the execution : " << endl;
      cerr << s << endl;
      exit(1);
   }
}
