/* misc.h */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
*/

/*!
 * \file
 *                Multiscale KSVD Denoising
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File misc.h
 * \brief Contains miscellaneous functions */

#ifndef MISC_H
#define MISC_H

#include <qimage.h>
#include <qstring.h>
#include <stdlib.h>
#include <iostream>
#include <assert.h>
#include <limits>
#include "cpplapack/cpplapack.h"
#include "tnt/tnt_sparse_matrix.h"

#define GAMMA 1.5
#define MAX_N 5
#define INFINITY 10e150
#define EPSILON 10e-6
#define SQRT2 1.4142
#define MAX_VALUE 255.0
#define MIN_VALUE 0.0
#define MEAN_VALUE 127.5
#define MAX(A,B) ((A>B)?A:B)
#define MIN(A,B) ((A<B)?A:B)

using namespace CPPL;
using namespace std;

typedef dgematrix Matrix;
typedef dcovector VectC;
typedef drovector VectR;
typedef dsymatrix Symmetric_Matrix;

static const int twoPow[MAX_N+1] = { 1, 2, 4, 8, 16, 32};
static const int fourPow[MAX_N+1] = { 1, 4, 16, 64, 256, 4096};
static int seed;

#ifndef USE_LAPACK
extern "C" {
   /* return the index of element having the smallest absolute value */
   long idamin_( const long &N, const double *x, const long &incx );
};
#else
long inline idamin_( const long &N, const double *x, const long &incx ) {
   double min = x[0];
   long index = 0;
   for (long i = 1; i<N; ++i) {
      if (min > x[i]) {
         min = x[i];
         index = i;
      }
   }
   return index;
};

#endif // USE_LAPACK

/*! \brief
 * K-SVD namespace
 * */
namespace KSVD {

   /// fonction only used for debugging (without a debugger :-) )
   static inline void stop() {
      cout << "Appuyez sur entre pour continuer...";
      cin.ignore( numeric_limits<streamsize>::max(), '\n' );
   };

   /// uniform distribution, taken from Numerical Recipes in C++
   static inline double ran1() {
      const int IA=16807,IM=2147483647,IQ=127773,IR=2836,NTAB=32;
      const int NDIV=(1+(IM-1)/NTAB);
      const double EPS=3.0e-16,AM=1.0/IM,RNMX=(1.0-EPS);
      static int iy=0;
      static int iv[NTAB];
      int j,k;
      double temp;

      if (seed <= 0 || !iy) {
         if (-seed < 1) seed=1;
         else seed = -seed;
         for (j=NTAB+7;j>=0;j--) {
            k=seed/IQ;
            seed=IA*(seed-k*IQ)-IR*k;
            if (seed < 0) seed += IM;
            if (j < NTAB) iv[j] = seed;
         }
         iy=iv[0];
      }
      k=seed/IQ;
      seed=IA*(seed-k*IQ)-IR*k;
      if (seed < 0) seed += IM;
      j=iy/NDIV;
      iy=iv[j];
      iv[j] = seed;
      if ((temp=AM*iy) > RNMX) return RNMX;
      else return temp;
   };

   /// generate a value from a normal distribution (taken in 
   /// Numerical Recipes in C++
   template <typename T>
      static inline T normalDistrib() {
         static bool iset = true;
         static T gset;
         /// the function rand is enough for our usage.
         /// can switch to ran1 if needed
         T fac,rsq,v1,v2;
         if (iset) {
            do {
               v1 = 2.0*ran1()-1.0;
               v2 = 2.0*ran1()-1.0;
               rsq = v1*v1+v2*v2;
            } while (rsq >= 1.0 || rsq == 0.0);
            fac = sqrt(-2.0*log(rsq)/rsq);
            gset = v1*fac;
            iset = false;
            return v2*fac;
         } else {
            iset = true;
            return gset;
         }
      };


   /// change the metric using the \f$ \gamma \f$ term
   /// can eventually work with V != 3
   template <typename T, int V>
      static inline void switchPatchToRGBMetric(VectR& patch, int level,
            const T a) {
         assert(level >= 0);
         if (level == 0) {
            VectR mean(static_cast<long>(V));
            for (int i = 0; i<V; ++i) mean(i) = 0;
            int channel = 0;
            for (int i = 0; i<patch.l; ++i) {
               mean(channel) += patch(i);
               channel = (channel + 1) % V;
            }
            T aa = V*a/patch.l;
            channel = 0;
            for (int i = 0; i<patch.l; ++i) {
               patch(i) += aa*mean(channel);
               channel = (channel + 1) % V;
            }
         } else {
            int newL = patch.l/4;
            for (int i = 0; i< 4; ++i) {
               VectR quad(newL,patch,i);
               switchPatchToRGBMetric<T,V>(quad,level-1,a);
               quad.setArrayNull();
            }
         }
      };

   /// change the metric using the \f$ \gamma \f$ term
   /// can eventually work with V != 3
   template <typename T, int V>
      static inline void switchPatchBackFromRGBMetric(VectR& patch, 
            int level,
            const T a) {
         switchPatchToRGBMetric<T,V>(patch,level,-a/(1.0+a));
      };

   /// change the metric of a dictionary using a \f$ \beta \f$ matrix
   /// and normalize the dictionary 
   template <typename T>
      static inline void applyWeightToDict(Matrix& dict, const VectR&
            beta) {
         for (int i = 0; i<dict.n; ++i) {
            for (int j = 0; j<dict.m; ++j) {
               dict(j,i) *= beta(j);
            }
         }
      };

   /// change the metric of a dictionary using a \f$ \beta \f$ matrix
   /// and normalize the dictionary 
   template <typename T, int V>
      static inline void switchDictToRGBMetric(Matrix& dict, 
            const int sizeUnit, const int numUnit, const T gamma) {
         for (int i = 0; i<dict.n; ++i) {
            VectR atom = dict.colRef(i);
            for (int j = 0; j< numUnit; ++j) {
               VectR atomQuad(sizeUnit,atom,j);
               switchPatchToRGBMetric<T,V>(atomQuad,0,gamma);
               atomQuad.setArrayNull();
            }
            atom.setArrayNull();
         }
      };

   /// normalize a dictionary while saving the inverse of the norm
   /// of each atom.
   template <typename T>
      static inline void normalizeDict(Matrix& dict, T* invNorm) {
         for (int i = 0; i<dict.n; ++i) {
            VectR atom = dict.colRef(i);
            T nr = nrm2(atom);
            if (nr > EPSILON) {
               invNorm[i] = 1.0/nrm2(atom);
               atom *= invNorm[i];
               atom.setArrayNull();    
            } else {
               invNorm[i] = 1.0;
               atom = 1/sqrt(static_cast<T>(atom.l));
               atom.setArrayNull();    
            }
         }
      };

   template <typename T>
      static void openKernel(const QString& fileKernel, Matrix& kernel) {
         int Nx, Ny;
         ifstream f(fileKernel);
         if (!f) {
            cerr << "Failed to open kernel file" << endl;
            exit(1);
         }
         int N,M;
         f >> N >> M;
         // TODO
         f.close();
      };

   template <typename T>
      static void openKernelSep(const QString& fileKernel, VectR& A,
            VectR& B) {
         ifstream f(fileKernel);
         if (!f) {
            cerr << "Failed to open kernel file" << endl;
            exit(1);
         }
         int mode;
         f >> mode;
         if (mode != 1) {
            cerr << "Call openKernel instead of openKernelSep" << endl;
            exit(1);
         }
         int N,M;
         f >> N >> M;
         A.resize(N);
         B.resize(M);
         for (int i = 0; i<N; ++i) f >> A(i);
         for (int i = 0; i<M; ++i) f >> B(i);
         f.close();
      };

   /// remove some percents of a list of patches, fact is the ratio of patches
   /// to keep
   static void pruneListPatches(std::list<int>& listPatches, const double fact) {
      list<int>::iterator posErase;
      bool toErase = false;
      for (list<int>::iterator it = listPatches.begin();
            it != listPatches.end();
            ++it) {
         if (toErase) {
            toErase = false;
            listPatches.erase(posErase);
         }
         if (ran1() > fact) {
            posErase = it;
            toErase  = true;
         }
      }
   }      
};

/*!
 * \brief namespace used for adding some functions to CPPLapack library
 * */
namespace CPPL {

   /// function used for the sparse reconstruction
   template <typename T>
      static inline void addMultSparse(const Matrix& dict,
            const TNT::Sparse_Vector<T>& alpha,
            VectR& patch) {
         if (alpha.num_nonzeros() != 0) {
            for (typename TNT::Sparse_Vector<T>::const_iterator
                  it = alpha.begin();
                  it != alpha.end();
                  ++it) {
               patch += it->value()*dict.col(it->index());
            }
         }
      };

   /// debugging function
   static inline bool check(const VectR& vec) {
      for (int i = 0; i<vec.l; ++i) {
         if (isnan(vec(i))) return false;
      }
      return true;
   };

   /// debugging function for a matrix
   static inline bool check(const Matrix& mat) {
      for (int i = 0; i<mat.m*mat.n; ++i) {
         if (isnan(mat.array[i])) return false;
      }
      return true;
   };

   /// perform a simple sum of a vector
   static inline double sum(const VectR& vec) {
      double res = 0.0;
      for (int i = 0; i<vec.l; ++i) res += vec(i);
      return res;
   };

   /// count the number of non-zero elements
   static inline int numNonZeros(const VectR& vec) {
      int count = 0;
      for (int i = 0; i<vec.l; ++i) 
         if (fabs(vec(i)) > EPSILON) ++count;
      return count;
   };

   /// tell if a patch is constant or not 
   static inline bool isConstant(const VectR& vec) {
      for (int i = 1; i<vec.l; ++i) 
         if (vec(i) != vec(i-1)) return false;
      return true;
   };

   /// scale and shift the values of a patch to be suitable
   /// for display
   static inline void normalizePatchDisplay(VectR& patch) {
      int N = patch.l;
      double min = INFINITY;
      double max = -INFINITY;
      for (int i = 0; i<N; ++i) {
         if (patch(i) < min) min = patch(i);
         if (patch(i) > max) max = patch(i);
      }
      if (min == max) {
         for (int i = 0; i<N; ++i) {
            patch(i) = 127.0;
         }
      } else {
         for (int i = 0; i<N; ++i) {
            patch(i) = floor((patch(i)-min)*MAX_VALUE/(max-min));
         }
      }
   };

   /// change the structure of one patch to follow the multiscale 
   /// scheme
   template <typename T, int V>
      static inline void changeStructurePatch(VectR& vec,
            int sizeEdge, int level) {
         assert(level >= 0);
         if (level == 0) return;
         VectR* tmp = new VectR(vec.l);
         int newL = vec.l/4;
         int newSizeEdge = sizeEdge/2;
         int count1,count2;
         count1 = count2 = 0;
         for (int i = 0; i<newSizeEdge; ++i) {
            for (int j = 0; j<V*newSizeEdge; ++j) {
               (*tmp)(count1) = vec(count1+count2);
               ++count1;
            }
            for (int j = 0; j<V*newSizeEdge; ++j) {
               (*tmp)(count2+newL) = vec(count1+count2);
               ++count2;
            }
         }
         count1 = count2 = 0;
         for (int i = 0; i<newSizeEdge; ++i) {
            for (int j = 0; j<V*newSizeEdge; ++j) {
               (*tmp)(count1+2*newL) = vec(count1+count2+2*newL);
               ++count1;
            }
            for (int j = 0; j<V*newSizeEdge; ++j) {
               (*tmp)(count2+3*newL) = vec(count1+count2+2*newL);
               ++count2;
            }
         }
         vec.copyPart(*tmp,0);
         delete(tmp);
         VectR* quad;
         for (int i = 0; i<4; ++i) {
            quad = new VectR(newL,vec,i);
            CPPL::changeStructurePatch<T,V>(*quad,newSizeEdge,level-1);
            quad->setArrayNull();
            delete(quad);
         }
      };

   /// reverse the structure of one patch
   template <typename T, int V>
      static inline void reverseStructurePatch(VectR& vec,
            int level) {
         if (level == 0) return;
         VectR* quad;
         int newL = vec.l/4;
         for (int i = 0; i<4; ++i) {
            quad = new VectR(newL,vec,i);
            CPPL::reverseStructurePatch<T,V>(*quad,level-1);
            quad->setArrayNull();
            delete(quad);
         }
         int newSizeEdge = static_cast<int>(std::sqrt(
                  static_cast<T>(newL/V)));
         VectR* tmp = new VectR(vec.l);
         int count1,count2;
         count1 = count2 = 0;
         for (int i = 0; i<newSizeEdge; ++i) {
            for (int j = 0; j<V*newSizeEdge; ++j) {
               (*tmp)(count1+count2) = vec(count1);
               ++count1;
            }
            for (int j = 0; j<V*newSizeEdge; ++j) {
               (*tmp)(count2+count1) = vec(newL+count2);
               ++count2;
            }
         }
         count1 = count2 = 0;
         for (int i = 0; i<newSizeEdge; ++i) {
            for (int j = 0; j<V*newSizeEdge; ++j) {
               (*tmp)(count1+count2+2*newL) = vec(count1+2*newL);
               ++count1;
            }
            for (int j = 0; j<V*newSizeEdge; ++j) {
               (*tmp)(count2+count1+2*newL) = vec(count2+3*newL);
               ++count2;
            }
         }
         vec.copyPart(*tmp,0);
         delete(tmp);
      };

};

#endif // MISC_H
